# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly Reporter"""
# pylint: disable=protected-access

import os
import sys
import tarfile
from pathlib import Path

import pytest
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo

from .reporter import (
    FilesystemReporter,
    FuzzManagerReporter,
    Report,
    Reporter,
    S3FuzzManagerReporter,
)
from .storage import TestCase


def _create_crash_log(log_path):
    with log_path.open("w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: SEGV on unknown address 0x0")
        log_fp.write(" (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19")


def test_report_01(tmp_path):
    """test Report() with boring logs (no stack)"""
    (tmp_path / "not_a_log.txt").touch()
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    report = Report(str(tmp_path), "a.bin", size_limit=0)
    assert report._target_binary == "a.bin"
    assert report.path == str(tmp_path)
    assert report._logs.aux is None
    assert report._logs.stderr.endswith("log_stderr.txt")
    assert report._logs.stdout.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_stderr.txt")
    assert report.stack is None
    assert Report.DEFAULT_MAJOR == report.major
    assert Report.DEFAULT_MINOR == report.minor
    assert report.prefix is not None
    report.cleanup()
    assert not tmp_path.exists()


def test_report_02(tmp_path):
    """test Report() with crash logs"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    report = Report(str(tmp_path), "bin")
    assert report.path == str(tmp_path)
    assert report._logs.aux.endswith("log_asan_blah.txt")
    assert report._logs.stderr.endswith("log_stderr.txt")
    assert report._logs.stdout.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_asan_blah.txt")
    assert report.stack is not None
    assert Report.DEFAULT_MAJOR != report.major
    assert Report.DEFAULT_MINOR != report.minor
    assert report.prefix is not None
    report.cleanup()


def test_report_03(tmp_path):
    """test Report.tail()"""
    tmp_file = tmp_path / "file.txt"
    tmp_file.write_bytes(b"blah\ntest\n123\xEF\x00FOO")
    length = tmp_file.stat().st_size
    # no size limit
    with pytest.raises(AssertionError):
        Report.tail(str(tmp_file), 0)
    assert tmp_file.stat().st_size == length
    Report.tail(str(tmp_file), 3)
    log_data = tmp_file.read_bytes()
    assert log_data.startswith(b"[LOG TAILED]\n")
    assert log_data[13:] == b"FOO"


def test_report_04(tmp_path):
    """test Report.select_logs() uninteresting data"""
    # test with empty path
    assert Report.select_logs(str(tmp_path)) is None
    # empty file
    (tmp_path / "not_a_log.txt").touch()
    assert Report.select_logs(str(tmp_path)) is None


def test_report_05(tmp_path):
    """test Report.select_logs()"""
    # small log with nothing interesting
    with (tmp_path / "log_asan.txt.1").open("wb") as log_fp:
        log_fp.write(b"SHORT LOG\n")
        log_fp.write(b"filler line")
    # crash on another thread
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        log_fp.write(b"GOOD LOG\n")
        log_fp.write(
            b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x00000BADF00D"
        )
        log_fp.write(
            b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T0)\n"
        )  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    # child log that should be ignored (created when parent crashes)
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(
            b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000"
        )
        log_fp.write(
            b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T2)\n"
        )  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    (tmp_path / "log_mindump_blah.txt").write_bytes(b"minidump log")
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    # should be ignored in favor of "GOOD LOG"
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert "GOOD LOG" in (tmp_path / log_map.aux).read_text()
    assert "STDERR" in (tmp_path / log_map.stderr).read_text()
    assert "STDOUT" in (tmp_path / log_map.stdout).read_text()


def test_report_06(tmp_path):
    """test minidump with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|SIGSEGV|0x0|0\n")
        log_fp.write(b"minidump log\n")
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map.stderr).is_file()
    assert (tmp_path / log_map.stdout).is_file()
    assert "minidump log" in (tmp_path / log_map.aux).read_text()


def test_report_07(tmp_path):
    """test selecting preferred DUMP_REQUESTED minidump with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:739484451a63|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:739484451a63|1644|0x12\n")
    with (tmp_path / "log_minidump_02.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f57ac9e2e14|0\n")
        log_fp.write(
            b"0|0|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|"
            b"bar.cc:234|674|0xc\n"
        )
        log_fp.write(
            b"0|1|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|"
            b"bar.cc:4a2|645|0x8\n"
        )
    with (tmp_path / "log_minidump_03.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:1234|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:4323|1644|0x12\n")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map.stderr).is_file()
    assert (tmp_path / log_map.stdout).is_file()
    assert (
        "google_breakpad::ExceptionHandler::WriteMinidump"
        in (tmp_path / log_map.aux).read_text()
    )


def test_report_08(tmp_path):
    """test selecting worker logs with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_ffp_worker_1.txt").write_bytes(b"worker log")
    # we should only ever see one but if we see multiple we warn, so test that.
    (tmp_path / "log_ffp_worker_2.txt").write_bytes(b"worker log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map.stderr).is_file()
    assert (tmp_path / log_map.stdout).is_file()
    assert "worker log" in (tmp_path / log_map.aux).read_text()


def test_report_09(tmp_path):
    """test prioritizing sanitizer logs with Report._find_sanitizer()"""
    # NOTE: ordered by selection priority in order to use previously added logs
    # test empty
    (tmp_path / "log_asan.txt.0").touch()
    assert Report._find_sanitizer([str(x) for x in tmp_path.iterdir()]) is None
    # test *San log with data
    (tmp_path / "log_asan.txt.1").write_text("test")
    selected = Report._find_sanitizer([str(x) for x in tmp_path.iterdir()])
    assert selected is not None
    assert "test" in Path(selected).read_text()
    # test UBSan log
    (tmp_path / "log_asan.txt.1").write_text(
        "test.cc:3:5: runtime error: signed integer overflow: ..."
    )
    selected = Report._find_sanitizer([str(x) for x in tmp_path.iterdir()])
    assert selected is not None
    assert "runtime error: signed integer overflow" in Path(selected).read_text()
    # test selecting ASan report
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        # missing stack
        log_fp.write(b"==1184==ERROR: AddressSanitizer: BUS on ... blah\n")
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"==9482==ERROR: AddressSanitizer: stack-overflow on ...\n")
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % (l_no,))
    selected = Report._find_sanitizer([str(x) for x in tmp_path.iterdir()])
    assert selected is not None
    assert "AddressSanitizer: stack-overflow" in Path(selected).read_text()
    # test selecting prioritized
    with (tmp_path / "log_asan.txt.4").open("wb") as log_fp:
        log_fp.write(
            b"==1942==ERROR: AddressSanitizer: heap-use-after-free on ... blah\n"
        )
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % (l_no,))
    with (tmp_path / "log_asan.txt.5").open("wb") as log_fp:
        log_fp.write(b"==1984==ERROR: AddressSanitizer: SEGV on ... blah\n")
        log_fp.write(b"missing trace...\n")
    with (tmp_path / "log_asan.txt.6").open("wb") as log_fp:
        log_fp.write(b"ERROR: Failed to mmap\n")
    selected = Report._find_sanitizer([str(x) for x in tmp_path.iterdir()])
    assert selected is not None
    assert "heap-use-after-free" in Path(selected).read_text()


def test_report_10(tmp_path):
    """test Report() size_limit"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log\n" * 200)
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log\n" * 200)
    (tmp_path / "unrelated.txt").write_bytes(b"nothing burger\n" * 200)
    (tmp_path / "rr-trace").mkdir()
    size_limit = len("STDERR log\n")
    report = Report(str(tmp_path), "bin", size_limit=size_limit)
    assert report.path == str(tmp_path)
    assert report._logs.aux is None
    assert report._logs.stderr.endswith("log_stderr.txt")
    assert report._logs.stdout.endswith("log_stdout.txt")
    assert report.preferred.endswith("log_stderr.txt")
    assert report.stack is None
    size_limit += len("[LOG TAILED]\n")
    assert os.stat(os.path.join(report.path, report._logs.stderr)).st_size == size_limit
    assert os.stat(os.path.join(report.path, report._logs.stdout)).st_size == size_limit
    assert os.stat(os.path.join(report.path, "unrelated.txt")).st_size == size_limit
    report.cleanup()
    assert not tmp_path.is_dir()


def test_report_11(tmp_path):
    """test selecting Valgrind logs with Report.select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_valgrind.txt").write_bytes(b"valgrind log")
    log_map = Report.select_logs(str(tmp_path))
    assert (tmp_path / log_map.stderr).is_file()
    assert (tmp_path / log_map.stdout).is_file()
    assert "valgrind log" in (tmp_path / log_map.aux).read_text()


def test_report_12(tmp_path):
    """test Report.crash_info"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    # no binary.fuzzmanagerconf
    report = Report(str(tmp_path), target_binary="fake_bin")
    assert report._crash_info is None
    assert report.crash_info is not None
    assert report._crash_info is not None
    # with binary.fuzzmanagerconf
    with (tmp_path / "fake_bin.fuzzmanagerconf").open("wb") as conf:
        conf.write(b"[Main]\n")
        conf.write(b"platform = x86-64\n")
        conf.write(b"product = mozilla-central\n")
        conf.write(b"os = linux\n")
    report = Report(str(tmp_path), target_binary=str(tmp_path / "fake_bin"))
    assert report._crash_info is None
    assert report.crash_info is not None
    assert report._crash_info is not None


def test_report_13(mocker, tmp_path):
    """test Report.crash_signature and Report.crash_hash"""
    mocker.patch("grizzly.common.reporter.ProgramConfiguration", autospec=True)
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    report = Report(str(tmp_path), "bin")
    assert report._signature is None
    assert report.crash_signature
    assert report.crash_info.createShortSignature() == "[@ foo]"
    assert report.crash_hash


def test_report_14(mocker):
    """test Report.crash_signature_max_frames()"""
    info = mocker.Mock(spec=CrashInfo)
    info.backtrace = ("blah",)
    assert Report.crash_signature_max_frames(info) == 8
    info.backtrace = (
        "std::panicking::rust_panic",
        "std::panicking::rust_panic_with_hook",
    )
    assert Report.crash_signature_max_frames(info) == 14


def test_reporter_01(mocker):
    """test creating a simple Reporter"""

    class SimpleReporter(Reporter):
        def _pre_submit(self, report):
            pass

        def _post_submit(self):
            pass

        def _submit_report(self, report, test_cases):
            pass

    reporter = SimpleReporter()
    reporter.submit([], report=mocker.Mock(spec=Report))


def test_filesystem_reporter_01(tmp_path):
    """test FilesystemReporter without testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    report_path = tmp_path / "reports"
    report_path.mkdir()
    reporter = FilesystemReporter(str(report_path))
    reporter.submit([], Report(str(log_path), "fake_bin"))
    buckets = tuple(report_path.iterdir())
    # check major bucket
    assert len(buckets) == 1
    assert buckets[0].is_dir()
    # check log path exists
    log_dirs = tuple(buckets[0].iterdir())
    assert len(log_dirs) == 1
    assert log_dirs[0].is_dir()
    assert "_logs" in str(log_dirs[0])


def test_filesystem_reporter_02(tmp_path, mocker):
    """test FilesystemReporter with testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(log_path / "log_asan_blah.txt")
    tests = list(mocker.Mock(spec=TestCase) for _ in range(10))
    report_path = tmp_path / "reports"
    assert not report_path.exists()
    reporter = FilesystemReporter(str(report_path))
    reporter.submit(tests, Report(str(log_path), "fake_bin"))
    assert not log_path.exists()
    assert report_path.exists()
    assert len(tuple(report_path.iterdir())) == 1
    assert all(x.dump.call_count == 1 for x in tests)
    # call report a 2nd time
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    tests = list(mocker.Mock(spec=TestCase) for _ in range(2))
    reporter.submit(tests, Report(str(log_path), "fake_bin"))
    assert all(x.dump.call_count == 1 for x in tests)
    assert len(tuple(report_path.iterdir())) == 2
    assert len(tuple(report_path.glob("NO_STACK"))) == 1


def test_filesystem_reporter_03(tmp_path):
    """test FilesystemReporter disk space failsafe"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    reporter = FilesystemReporter(str(tmp_path / "reports"))
    reporter.min_space = 2 ** 50
    with pytest.raises(RuntimeError, match="Running low on disk space"):
        reporter.submit([], Report(str(log_path), "fake_bin"))


def test_filesystem_reporter_04(mocker, tmp_path):
    """test FilesystemReporter w/o major bucket"""
    fake_report = tmp_path / "fake_report"
    fake_report.mkdir()
    report = mocker.Mock(spec=Report, path=str(fake_report), prefix="test_prefix")
    reporter = FilesystemReporter(str(tmp_path / "dst"), major_bucket=False)
    reporter.submit([], report)
    assert not fake_report.is_dir()
    assert not report.major.call_count
    assert any((tmp_path / "dst").glob("test_prefix_logs"))


def test_fuzzmanager_reporter_01(mocker, tmp_path):
    """test FuzzManagerReporter.sanity_check()"""
    fake_reporter = mocker.patch("grizzly.common.reporter.ProgramConfiguration")
    fake_reporter.fromBinary.return_value = mocker.Mock(spec=ProgramConfiguration)
    # missing global FM config file
    FuzzManagerReporter.FM_CONFIG = "no_file"
    with pytest.raises(IOError, match="Missing: no_file"):
        FuzzManagerReporter.sanity_check("fake")
    # missing binary FM config file
    fake_fmc = tmp_path / ".fuzzmanagerconf"
    fake_fmc.touch()
    fake_bin = tmp_path / "bin"
    fake_bin.touch()
    FuzzManagerReporter.FM_CONFIG = str(fake_fmc)
    with pytest.raises(IOError, match="bin.fuzzmanagerconf"):
        FuzzManagerReporter.sanity_check(str(fake_bin))
    # success
    (tmp_path / "bin.fuzzmanagerconf").touch()
    FuzzManagerReporter.sanity_check(str(fake_bin))
    assert fake_reporter.fromBinary.call_count == 1


def test_fuzzmanager_reporter_02(mocker, tmp_path):
    """test FuzzManagerReporter.submit()"""
    mocker.patch(
        "grizzly.common.reporter.getcwd", autospec=True, return_value=str(tmp_path)
    )
    mocker.patch("grizzly.common.reporter.getenv", autospec=True, return_value="0")
    fake_crashinfo = mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_crashinfo.fromRawCrashData.return_value.createShortSignature.return_value = (
        "test [@ test]"
    )
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = str(
        tmp_path / "fm_file.signature"
    )
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_ffp_worker_blah.txt").touch()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    (log_path / "rr-traces").mkdir()
    (tmp_path / "screenlog.0").touch()
    fake_test = mocker.Mock(
        spec=TestCase,
        adapter_name="adapter",
        env_vars={"TEST": "1"},
        input_fname="input",
    )
    reporter = FuzzManagerReporter("fake_bin")
    reporter.submit([fake_test], Report(str(log_path), "fake_bin", is_hang=True))
    assert not log_path.is_dir()
    assert fake_test.dump.call_count == 1
    assert fake_collector.return_value.submit.call_count == 1
    meta_data = (tmp_path / "fm_file.metadata").read_text()
    assert '"frequent": false' in meta_data
    assert '"_grizzly_seen_count": 1' in meta_data
    assert '"shortDescription": "test [@ test]"' in meta_data


def test_fuzzmanager_reporter_03(mocker, tmp_path):
    """test FuzzManagerReporter.submit() - no test / mark as frequent"""
    mocker.patch(
        "grizzly.common.reporter.getcwd", autospec=True, return_value=str(tmp_path)
    )
    fake_crashinfo = mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_crashinfo.fromRawCrashData.return_value.createShortSignature.return_value = (
        "test [@ test]"
    )
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = str(
        tmp_path / "fm_file.signature"
    )
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter = FuzzManagerReporter("fake_bin")
    reporter.max_reports = 1
    reporter.submit([], Report(str(log_path), "fake_bin"))
    assert fake_collector.return_value.submit.call_count == 1
    meta_data = (tmp_path / "fm_file.metadata").read_text()
    assert '"frequent": true' in meta_data
    assert '"_grizzly_seen_count": 1' in meta_data
    assert '"shortDescription": "test [@ test]"' in meta_data


def test_fuzzmanager_reporter_04(mocker, tmp_path):
    """test FuzzManagerReporter.submit() hit frequent crash"""
    mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (
        None,
        {"frequent": True, "shortDescription": "[@ test]"},
    )
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter = FuzzManagerReporter("fake_bin")
    reporter.submit([], Report(str(log_path), "fake_bin"))
    fake_collector.return_value.submit.assert_not_called()


def test_fuzzmanager_reporter_05(mocker, tmp_path):
    """test FuzzManagerReporter.submit() hit existing crash"""
    mocker.patch("grizzly.common.reporter.CrashInfo", autospec=True)
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (
        None,
        {"bug__id": 1, "frequent": False, "shortDescription": "[@ test]"},
    )
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter = FuzzManagerReporter("fake_bin")
    reporter._ignored = lambda x: True
    reporter.submit([], Report(str(log_path), "fake_bin"))
    fake_collector.return_value.submit.assert_not_called()


def test_fuzzmanager_reporter_06(mocker, tmp_path):
    """test FuzzManagerReporter.submit() no signature"""
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = None
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter = FuzzManagerReporter("fake_bin")
    with pytest.raises(RuntimeError, match="Failed to create FM signature"):
        reporter.submit([], Report(str(log_path), "fake_bin"))


def test_fuzzmanager_reporter_07(mocker, tmp_path):
    """test FuzzManagerReporter.submit() unsymbolized crash"""
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (None, None)
    fake_collector.return_value.generate.return_value = None
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    reporter = FuzzManagerReporter("fake_bin")
    reporter._ignored = lambda x: True
    reporter.submit([], Report(str(log_path), "fake_bin"))
    fake_collector.return_value.submit.assert_not_called()


def test_fuzzmanager_reporter_08():
    """test FuzzManagerReporter.quality_name()"""
    assert FuzzManagerReporter.quality_name(0) == "QUAL_REDUCED_RESULT"
    assert FuzzManagerReporter.quality_name(-1) == "unknown quality (-1)"


def test_fuzzmanager_reporter_09(mocker, tmp_path):
    """test FuzzManagerReporter._ignored()"""
    log_file = tmp_path / "test.log"
    log_file.touch()
    report = mocker.Mock(spec=Report, path=str(tmp_path), preferred=str(log_file))
    # not ignored
    assert not FuzzManagerReporter._ignored(report)
    # ignored - sanitizer OOM missing stack
    log_file.write_bytes(b"ERROR: Failed to mmap")
    assert FuzzManagerReporter._ignored(report)
    # ignored - Valgrind OOM
    log_file.write_bytes(b"VEX temporary storage exhausted.")
    assert FuzzManagerReporter._ignored(report)


def test_s3fuzzmanager_reporter_01(mocker, tmp_path):
    """test S3FuzzManagerReporter.sanity_check()"""
    mocker.patch("grizzly.common.reporter.FuzzManagerReporter", autospec=True)
    fake_bin = tmp_path / "bin"
    # test GRZ_S3_BUCKET missing
    with pytest.raises(
        EnvironmentError, match="'GRZ_S3_BUCKET' is not set in environment"
    ):
        S3FuzzManagerReporter.sanity_check(str(fake_bin))
    # test GRZ_S3_BUCKET set
    pytest.importorskip("boto3")
    mocker.patch("grizzly.common.reporter.getenv", autospec=True, return_value="test")
    S3FuzzManagerReporter.sanity_check(str(fake_bin))


def test_s3fuzzmanager_reporter_02(mocker, tmp_path):
    """test S3FuzzManagerReporter._pre_submit()"""
    pytest.importorskip("boto3")
    pytest.importorskip("botocore")
    mocker.patch("grizzly.common.reporter.getenv", autospec=True, return_value="test")
    fake_resource = mocker.patch("grizzly.common.reporter.resource", autospec=True)

    fake_report = mocker.Mock(spec=Report)
    fake_report.path = "no-path"
    reporter = S3FuzzManagerReporter("fake_bin")
    # test will missing rr-trace
    assert reporter._pre_submit(fake_report) is None
    assert not reporter._extra_metadata

    # test will exiting rr-trace
    trace_dir = tmp_path / "rr-traces" / "latest-trace"
    trace_dir.mkdir(parents=True)
    fake_report.minor = "1234abcd"
    fake_report.path = str(tmp_path)
    reporter._pre_submit(fake_report)
    assert not any(tmp_path.iterdir())
    assert "rr-trace" in reporter._extra_metadata
    assert fake_report.minor in reporter._extra_metadata["rr-trace"]
    fake_resource.return_value.meta.client.upload_file.assert_not_called()

    # test with new rr-trace
    reporter._extra_metadata.clear()
    trace_dir.mkdir(parents=True)
    (trace_dir / "trace-file").touch()

    class FakeClientError(Exception):
        def __init__(self, message, response):
            super().__init__(message)
            self.response = response

    mocker.patch("grizzly.common.reporter.ClientError", new=FakeClientError)
    fake_resource.return_value.Object.side_effect = FakeClientError(
        "test", {"Error": {"Code": "404"}}
    )
    reporter._pre_submit(fake_report)
    assert not any(tmp_path.iterdir())
    assert "rr-trace" in reporter._extra_metadata
    assert fake_report.minor in reporter._extra_metadata["rr-trace"]
    assert fake_resource.return_value.meta.client.upload_file.call_count == 1


@pytest.mark.skipif(
    not sys.platform.startswith("linux"), reason="RR only supported on Linux"
)
def test_s3fuzzmanager_reporter_03(tmp_path):
    """test S3FuzzManagerReporter.compress_rr_trace()"""
    # create fake trace
    src = tmp_path / "rr-traces" / "echo-0"
    src.mkdir(parents=True)
    (src / "fail_file").touch()
    src = tmp_path / "rr-traces" / "echo-1"
    src.mkdir()
    (src / "cloned_data_5799_1").touch()
    (src / "data").write_bytes(b"test_data")
    (src / "events").write_bytes(b"foo")
    (src / "mmap").write_bytes(b"bar")
    (src / "tasks").write_bytes(b"foo")
    (src / "version").write_bytes(b"123")
    (tmp_path / "rr-traces" / "latest-trace").symlink_to(
        str(src), target_is_directory=True
    )
    src = tmp_path / "rr-traces"
    dest = tmp_path / "dest"
    dest.mkdir()
    S3FuzzManagerReporter.compress_rr_trace(str(src), str(dest))
    assert not src.is_dir()
    assert (dest / "rr.tar.bz2").is_file()
    with tarfile.open(str(dest / "rr.tar.bz2"), "r:bz2") as arc_fp:
        entries = arc_fp.getnames()
    assert "echo-1" in entries
    assert "echo-0" not in entries
    assert "latest-trace" not in entries


# TODO: fill out tests for FuzzManagerReporter and S3FuzzManagerReporter



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
from collections import namedtuple
from itertools import chain
from os import SEEK_END, listdir, makedirs, walk
from os.path import abspath, basename, dirname, isdir, isfile
from os.path import join as pathjoin
from os.path import normpath, relpath
from shutil import copyfileobj, rmtree
from tempfile import SpooledTemporaryFile, mkdtemp
from time import time
from zipfile import BadZipfile, ZipFile
from zlib import error as zlib_error

from .utils import grz_tmp, sanitizer_opts

__all__ = ("TestCase", "TestFile", "TestCaseLoadFailure", "TestFileExists")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class TestCaseLoadFailure(Exception):
    """Raised when loading a TestCase fails"""


class TestFileExists(Exception):
    """Raised when adding a TestFile to a TestCase that has an existing
    TestFile with the same name"""


TestFileMap = namedtuple("TestFileMap", "meta optional required")


class TestCase:
    __slots__ = (
        "adapter_name",
        "duration",
        "env_vars",
        "hang",
        "input_fname",
        "landing_page",
        "redirect_page",
        "time_limit",
        "timestamp",
        "_existing_paths",
        "_files",
    )

    def __init__(
        self,
        landing_page,
        redirect_page,
        adapter_name,
        input_fname=None,
        time_limit=None,
        timestamp=None,
    ):
        self.adapter_name = adapter_name
        self.duration = None
        self.env_vars = dict()  # environment variables
        self.hang = False
        self.input_fname = input_fname  # file that was used to create the test case
        self.landing_page = landing_page
        self.redirect_page = redirect_page
        self.time_limit = time_limit
        self.timestamp = time() if timestamp is None else timestamp
        self._existing_paths = list()  # file paths in use
        self._files = TestFileMap(
            meta=list(),  # environment files such as prefs.js, etc...
            optional=list(),
            required=list(),
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def _add(self, target, test_file):
        """Add a test file to test case and perform sanity checks.

        Args:
            target (list): Specific list of files to append target test_file to.
            test_file (TestFile): TestFile to append.

        Returns:
            None
        """
        assert isinstance(test_file, TestFile), "only accepts TestFiles"
        if test_file.file_name in self._existing_paths:
            raise TestFileExists("%r exists in test" % (test_file.file_name,))
        self._existing_paths.append(test_file.file_name)
        target.append(test_file)

    def add_batch(self, path, include_files, prefix=None):
        """Iterate over files in include_files and attach the files that are
        located in path to TestCase.

        Args:
            path (str): Path to the root of the directory that contains files.
            include_files (iterable(str)): Paths of the files to be added to the
                                           TestCase if they exist in path.
            prefix (str): Path prefix to prepend to file when adding to
                          the TestCase.

        Returns:
            None
        """
        path = abspath(path)
        for fname in (x for x in include_files if x.startswith(path)):
            test_path = relpath(fname, path)
            if test_path.startswith(".."):
                continue
            if prefix:
                test_path = "/".join((prefix, test_path))
            self.add_from_file(fname, file_name=test_path)

    def add_meta(self, meta_file):
        """Add a TestFile to TestCase as a meta file.

        Args:
            meta_file (TestFile): TestFile to add to TestCase.

        Returns:
            None
        """
        self._add(self._files.meta, meta_file)

    def add_environ_var(self, name, value):
        """Add environment variable to TestCase.

        Args:
            name (str): Environment variable name.
            value (str): Environment variable value.

        Returns:
            None
        """
        self.env_vars[name] = value

    def add_file(self, test_file, required=True):
        """Add a TestFile to TestCase.

        Args:
            meta_file (TestFile): TestFile to add to TestCase.
            required (bool): Indicates if test file must be served.

        Returns:
            None
        """
        if required:
            self._add(self._files.required, test_file)
        else:
            self._add(self._files.optional, test_file)

    def add_from_data(self, data, file_name, encoding="UTF-8", required=True):
        """Create a TestFile and add it to the TestCase.

        Args:
            data (bytes or str): Data to write to file. If data is of type str
                                 encoding must be given.
            file_name (str): Name for the TestFile.
            encoding (str): Encoding to be used.
            required (bool): Indicates whether the TestFile must be served.

        Returns:
            None
        """
        tfile = TestFile.from_data(data, file_name, encoding=encoding)
        try:
            self.add_file(tfile, required=required)
        except TestFileExists:
            tfile.close()
            raise

    def add_from_file(self, input_file, file_name=None, required=True):
        """Create a TestFile from an existing file and add it to the TestCase.

        Args:
            input_file (str): Path to existing file to use.
            file_name (str): Name for the TestFile. If file_name is not given
                             the name of the input_file will be used.
            required (bool): Indicates whether the TestFile must be served.

        Returns:
            None
        """
        tfile = TestFile.from_file(input_file, file_name=file_name)
        try:
            self.add_file(tfile, required=required)
        except TestFileExists:
            tfile.close()
            raise

    def cleanup(self):
        """Close all the test files.

        Args:
            None

        Returns:
            None
        """
        for file_group in self._files:
            for test_file in file_group:
                test_file.close()

    def clone(self):
        """Make a copy of the TestCase.

        Args:
            None

        Returns:
            TestCase: A copy of the TestCase instance.
        """
        result = type(self)(
            self.landing_page,
            self.redirect_page,
            self.adapter_name,
            self.input_fname,
            self.time_limit,
            self.timestamp,
        )
        result.duration = self.duration
        result.hang = self.hang
        result.env_vars.update(self.env_vars)
        for entry in self._files.meta:
            result.add_meta(entry.clone())
        for entry in self._files.optional:
            result.add_file(entry.clone(), required=False)
        for entry in self._files.required:
            result.add_file(entry.clone(), required=True)
        return result

    def contains(self, file_name):
        """Check TestCase contains the TestFile with name matching `file_name`.

        Args:
            file_name (str): File name to search for in TestCase.

        Returns:
            bool: True if file exists in the TestCase otherwise False.
        """
        return file_name in self._existing_paths

    @property
    def data_size(self):
        """The total amount of data used (bytes) by the TestFiles in the
        TestCase.

        Args:
            None

        Returns:
            int: Total size of the test case in byte.
        """
        total = 0
        for group in self._files:
            total += sum(x.size for x in group)
        return total

    def dump(self, out_path, include_details=False):
        """Write all the test case data to the filesystem.

        Args:
            out_path (str): Path to directory to output data.
            include_details (bool): Output "test_info.json" file.

        Returns:
            None
        """
        # save test files to out_path
        for test_file in chain(self._files.required, self._files.optional):
            test_file.dump(out_path)
        # save test case files and meta data including:
        # adapter used, input file, environment info and files
        if include_details:
            assert isinstance(self.env_vars, dict)
            info = {
                "adapter": self.adapter_name,
                "duration": self.duration,
                "env": self.env_vars,
                "hang": self.hang,
                "input": basename(self.input_fname) if self.input_fname else None,
                "target": self.landing_page,
                "time_limit": self.time_limit,
                "timestamp": self.timestamp,
            }
            with open(pathjoin(out_path, "test_info.json"), "w") as out_fp:
                json.dump(info, out_fp, indent=2, sort_keys=True)
            # save meta files
            for meta_file in self._files.meta:
                meta_file.dump(out_path)

    def get_file(self, file_name):
        """Lookup and return the TestFile with the specified file name.

        Args:
            file_name (str): Name of file to retrieve.

        Returns:
            TestFile: TestFile with matching file name otherwise None.
        """
        for tfile in chain(
            self._files.meta, self._files.optional, self._files.required
        ):
            if tfile.file_name == file_name:
                return tfile
        return None

    @classmethod
    def load(cls, path, load_prefs, adjacent=False):
        """Load TestCases from disk.

        Args:
            path (str): Path can be:
                        1) A directory containing `test_info.json` and data.
                        2) A directory with one or more subdirectories of 1.
                        3) A zip archive containing testcase data or
                           subdirectories containing testcase data.
                        4) A single file to be used as a test case.
            load_prefs (bool): Load prefs.js file if available.
            adjacent (bool): Load adjacent files as part of the test case.
                             This is always the case when loading a directory.
                             WARNING: This should be used with caution!

        Returns:
            list: TestCases successfully loaded from path.
        """
        # unpack archive if needed
        if path.lower().endswith(".zip"):
            unpacked = mkdtemp(prefix="unpack_", dir=grz_tmp("storage"))
            try:
                with ZipFile(path) as zip_fp:
                    zip_fp.extractall(path=unpacked)
            except (BadZipfile, zlib_error):
                rmtree(unpacked, ignore_errors=True)
                raise TestCaseLoadFailure("Testcase archive is corrupted") from None
            path = unpacked
        else:
            unpacked = None
        # load testcase data from disk
        try:
            if isfile(path):
                tests = [cls.load_single(path, load_prefs, adjacent=adjacent)]
            elif isdir(path):
                tests = list()
                for tc_path in TestCase.scan_path(path):
                    tests.append(cls.load_single(tc_path, load_prefs))
                tests.sort(key=lambda tc: tc.timestamp)
            else:
                raise TestCaseLoadFailure("Invalid TestCase path")
        finally:
            if unpacked is not None:
                rmtree(unpacked, ignore_errors=True)
        return tests

    def load_environ(self, path, env_data):
        # sanity check environment variable data
        for name, value in env_data.items():
            if not isinstance(name, str) or not isinstance(value, str):
                raise TestCaseLoadFailure("'env_data' contains invalid 'env' entries")
        self.env_vars = env_data
        known_suppressions = ("lsan.supp", "tsan.supp", "ubsan.supp")
        for supp in listdir(path):
            if supp.lower() in known_suppressions:
                # Update *SAN_OPTIONS to use provided suppression files.
                opt_key = "%s_OPTIONS" % (supp.split(".")[0].upper(),)
                opts = sanitizer_opts(self.env_vars.get(opt_key, ""))
                opts["suppressions"] = "'%s'" % (pathjoin(path, supp),)
                self.env_vars[opt_key] = ":".join(
                    "=".join((k, v)) for k, v in opts.items()
                )

    @classmethod
    def load_single(cls, path, load_prefs, adjacent=False):
        """Load contents of a TestCase from disk. If `path` is a directory it must
        contain a valid 'test_info.json' file.

        Args:
            path (str): Path to the directory or file to load.
            load_prefs (bool): Load prefs.js file if available.
            adjacent (bool): Load adjacent files as part of the TestCase.
                             This is always true when loading a directory.
                             WARNING: This should be used with caution!

        Returns:
            TestCase: A TestCase.
        """
        path = abspath(path)
        if isdir(path):
            # load using test_info.json
            try:
                with open(pathjoin(path, "test_info.json"), "r") as in_fp:
                    info = json.load(in_fp)
            except IOError:
                raise TestCaseLoadFailure("Missing 'test_info.json'") from None
            except ValueError:
                raise TestCaseLoadFailure("Invalid 'test_info.json'") from None
            if not isinstance(info.get("target"), str):
                raise TestCaseLoadFailure("'test_info.json' has invalid 'target' entry")
            entry_point = basename(info["target"])
            if not isfile(pathjoin(path, entry_point)):
                raise TestCaseLoadFailure(
                    "Entry point %r not found in '%s'" % (entry_point, path)
                )
            # always load all contents of a directory if a 'test_info.json' is loaded
            adjacent = True
        elif isfile(path):
            entry_point = basename(path)
            info = dict()
            path = dirname(path)
        else:
            raise TestCaseLoadFailure("Missing or invalid TestCase %r" % (path,))
        # create testcase and add data
        test = cls(
            None,
            None,
            info.get("adapter", None),
            input_fname=info.get("input", None),
            time_limit=info.get("time_limit", None),
            timestamp=info.get("timestamp", 0),
        )
        test.duration = info.get("duration", None)
        test.hang = info.get("hang", False)
        if load_prefs and isfile(pathjoin(path, "prefs.js")):
            test.add_meta(TestFile.from_file(pathjoin(path, "prefs.js")))
        test.add_from_file(pathjoin(path, entry_point))
        test.landing_page = entry_point
        # load environment variables
        if info:
            try:
                test.load_environ(path, info.get("env", {}))
            except TestCaseLoadFailure:
                test.cleanup()
                raise
        # load all adjacent data from directory
        if adjacent:
            for dpath, _, files in walk(path):
                for fname in files:
                    # ignore files that have been previously loaded
                    if fname in (entry_point, "prefs.js", "test_info.json"):
                        continue
                    location = "/".join((dpath.split(path, 1)[-1], fname))
                    test.add_from_file(
                        pathjoin(dpath, fname), file_name=location, required=False
                    )
        return test

    @property
    def optional(self):
        """Get file names of optional TestFiles.

        Args:
            None

        Yields:
            str: File names of optional files.
        """
        for test in self._files.optional:
            yield test.file_name

    def purge_optional(self, keep):
        """Remove optional files (by name) that are not in keep.

        Args:
            keep (iterable(str)): Files that will not be removed.

        Returns:
            None
        """
        opt_files = tuple(x.file_name for x in self._files.optional)
        if not opt_files:
            # nothing to purge
            return
        # filter required files from opt_files files to keep
        keep_opt = list()
        for fname in set(keep):
            if fname not in (x.file_name for x in self._files.required):
                keep_opt.append(fname)
        # sanity check keep (cannot remove file that does not exist)
        assert all(fname in opt_files for fname in keep_opt)
        # purge
        to_remove = list()
        for idx, fname in enumerate(opt_files):
            if fname not in keep_opt:
                to_remove.append(idx)
        for idx in reversed(to_remove):
            self._files.optional.pop(idx).close()

    @staticmethod
    def scan_path(path):
        """Check path and subdirectories for potential test cases.

        Args:
            path (str): Path to scan.

        Yields:
            str: Path to what appears to be a valid testcase.
        """
        contents = listdir(path)
        if "test_info.json" in contents:
            yield path
        else:
            for entry in contents:
                tc_path = pathjoin(path, entry)
                if isfile(pathjoin(tc_path, "test_info.json")):
                    yield tc_path


class TestFile:
    CACHE_LIMIT = 0x80000  # data cache limit per file: 512KB
    XFER_BUF = 0x10000  # transfer buffer size: 64KB

    __slots__ = ("_file_name", "_fp")

    def __init__(self, file_name):
        # This is a naive fix for a larger path issue. This is a simple sanity
        # check and does not check if invalid characters are used. If an invalid
        # file name is used an exception will be raised when trying to write
        # that file to the file system.
        if "\\" in file_name:
            file_name = file_name.replace("\\", "/")
        if file_name.startswith("/"):
            file_name = file_name.lstrip("/")
        if file_name.endswith("."):
            file_name = file_name.rstrip(".")
        if (
            not file_name
            or ("/" in file_name and not file_name.rsplit("/", 1)[-1])
            or file_name.startswith("../")
        ):
            raise TypeError("file_name is invalid %r" % (file_name,))
        # name including path relative to wwwroot
        self._file_name = normpath(file_name)
        # pylint: disable=consider-using-with
        self._fp = SpooledTemporaryFile(
            dir=grz_tmp("storage"), max_size=self.CACHE_LIMIT, prefix="testfile_"
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def clone(self):
        """Make a copy of the TestFile.

        Args:
            None

        Returns:
            TestFile: A copy of the TestFile instance
        """
        cloned = type(self)(self._file_name)
        self._fp.seek(0)
        copyfileobj(
            self._fp, cloned._fp, self.XFER_BUF  # pylint: disable=protected-access
        )
        return cloned

    def close(self):
        """Close the TestFile.

        Args:
            None

        Returns:
            None
        """
        self._fp.close()

    @property
    def data(self):
        """Get the data from the TestFile. Not recommenced for large files.

        Args:
            None

        Returns:
            bytes: Data from the TestFile
        """
        pos = self._fp.tell()
        self._fp.seek(0)
        data = self._fp.read()
        self._fp.seek(pos)
        return data

    def dump(self, path):
        """Write TestFile data to the filesystem.

        Args:
            path (str): Path to output data.

        Returns:
            None
        """
        target_path = pathjoin(path, dirname(self._file_name))
        if not isdir(target_path):
            makedirs(target_path)
        self._fp.seek(0)
        with open(pathjoin(path, self._file_name), "wb") as dst_fp:
            copyfileobj(self._fp, dst_fp, self.XFER_BUF)

    @property
    def file_name(self):
        return self._file_name

    @classmethod
    def from_data(cls, data, file_name, encoding="UTF-8"):
        """Create a TestFile and add it to the test case.

        Args:
            data (bytes or str): Data to write to file. If data is of type str
                                 encoding must be given.
            file_name (str): Name for the TestFile.
            encoding (str): Encoding to be used.

        Returns:
            TestFile: A TestFile.
        """
        t_file = cls(file_name)
        if data:
            if isinstance(data, bytes) or not encoding:
                t_file.write(data)
            else:
                t_file.write(data.encode(encoding))
        return t_file

    @classmethod
    def from_file(cls, input_file, file_name=None):
        """Create a TestFile from an existing file.

        Args:
            input_file (str): Path to existing file to use.
            file_name (str): Name for the TestFile. If file_name is not given
                             the name of the input_file will be used.

        Returns:
            TestFile: A TestFile.
        """
        if file_name is None:
            file_name = basename(input_file)
        t_file = cls(file_name)
        with open(input_file, "rb") as src_fp:
            copyfileobj(
                src_fp, t_file._fp, cls.XFER_BUF  # pylint: disable=protected-access
            )
        return t_file

    @property
    def size(self):
        """Size of the file in bytes.

        Args:
            None

        Returns:
            int: Size in bytes.
        """
        pos = self._fp.tell()
        self._fp.seek(0, SEEK_END)
        size = self._fp.tell()
        self._fp.seek(pos)
        return size

    def write(self, data):
        """Add data to the TestFile.

        Args:
            data (bytes): Data to add to the TestFile.

        Returns:
            None
        """
        self._fp.write(data)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from os import close, kill, unlink
from os.path import abspath, isfile
from platform import system
from signal import SIGABRT

try:
    from signal import SIGUSR1
except ImportError:
    SIGUSR1 = None
from tempfile import mkdtemp, mkstemp
from time import sleep, time

from ffpuppet import BrowserTimeoutError, FFPuppet, LaunchError
from prefpicker import PrefPicker
from psutil import AccessDenied, NoSuchProcess, Process, process_iter

from ..common.reporter import Report
from ..common.utils import grz_tmp
from .target import Target, TargetError, TargetLaunchError, TargetLaunchTimeout
from .target_monitor import TargetMonitor

__all__ = ("PuppetTarget",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


class PuppetTarget(Target):
    __slots__ = ("use_rr", "use_valgrind", "_puppet", "_remove_prefs")

    def __init__(
        self, binary, extension, launch_timeout, log_limit, memory_limit, **kwds
    ):
        super().__init__(binary, extension, launch_timeout, log_limit, memory_limit)
        self.use_rr = kwds.pop("rr", False)
        self.use_valgrind = kwds.pop("valgrind", False)
        self._remove_prefs = False
        # create Puppet object
        self._puppet = FFPuppet(
            use_rr=self.use_rr,
            use_valgrind=self.use_valgrind,
            use_xvfb=kwds.pop("xvfb", False),
        )
        if kwds:
            LOG.warning(
                "PuppetTarget ignoring unsupported arguments: %s", ", ".join(kwds)
            )

    def add_abort_token(self, token):
        self._puppet.add_abort_token(token)

    def cleanup(self):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.clean_up()
        if self._remove_prefs and self._prefs and isfile(self._prefs):
            unlink(self._prefs)

    def close(self):
        # prevent parallel calls to FFPuppet.close() and/or FFPuppet.clean_up()
        with self._lock:
            self._puppet.close()

    @property
    def closed(self):
        return self._puppet.reason is not None

    def is_idle(self, threshold):
        for _, cpu in self._puppet.cpu_usage():
            if cpu >= threshold:
                return False
        return True

    def create_report(self, is_hang=False):
        logs = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
        self.save_logs(logs)
        return Report(logs, self.binary, is_hang=is_hang)

    @property
    def monitor(self):
        if self._monitor is None:

            class _PuppetMonitor(TargetMonitor):
                # pylint: disable=no-self-argument,protected-access
                def clone_log(_, log_id, offset=0):
                    return self._puppet.clone_log(log_id, offset=offset)

                def is_running(_):
                    return self._puppet.is_running()

                def is_healthy(_):
                    return self._puppet.is_healthy()

                @property
                def launches(_):
                    return self._puppet.launches

                def log_length(_, log_id):
                    return self._puppet.log_length(log_id)

            self._monitor = _PuppetMonitor()
        return self._monitor

    def detect_failure(self, ignored):
        status = self.RESULT_NONE
        # check if there has been a crash, hangs will appear as SIGABRT
        if not self._puppet.is_healthy():
            self.close()
            # something has happened figure out what
            if self._puppet.reason == FFPuppet.RC_CLOSED:
                LOG.debug("target.close() was called")
            elif self._puppet.reason == FFPuppet.RC_EXITED:
                LOG.debug("target closed itself")
            elif (
                self._puppet.reason == FFPuppet.RC_WORKER
                and "memory" in ignored
                and "ffp_worker_memory_usage" in self._puppet.available_logs()
            ):
                status = self.RESULT_IGNORED
                LOG.debug("memory limit exceeded")
            elif (
                self._puppet.reason == FFPuppet.RC_WORKER
                and "log-limit" in ignored
                and "ffp_worker_log_size" in self._puppet.available_logs()
            ):
                status = self.RESULT_IGNORED
                LOG.debug("log size limit exceeded")
            else:
                # crash or hang (forced SIGABRT) has been detected
                LOG.debug("failure detected, ffpuppet reason %r", self._puppet.reason)
                status = self.RESULT_FAILURE
        return status

    def handle_hang(self, ignore_idle=True):
        was_idle = False
        if self._puppet.is_healthy():
            proc_usage = sorted(self._puppet.cpu_usage(), key=lambda x: x[1])
            if proc_usage:
                pid, cpu = proc_usage.pop()
                if ignore_idle and cpu < 15:
                    # don't send SIGABRT if process is idle
                    LOG.debug("ignoring idle hang (%0.1f%%)", cpu)
                    was_idle = True
                elif system() == "Linux":
                    # sending SIGABRT is only supported on Linux for now
                    # TODO: add/test on other OSs
                    LOG.debug("sending SIGABRT to %r (%0.1f%%)", pid, cpu)
                    try:
                        kill(pid, SIGABRT)
                    except OSError:
                        LOG.warning("Failed to send SIGABRT to pid %d", pid)
                    self._puppet.wait(timeout=10)
        # always call close() since this function should only/always
        # be called when there has been a timeout
        self.close()
        return was_idle

    def dump_coverage(self, timeout=15):
        assert SIGUSR1 is not None
        pid = self._puppet.get_pid()
        if pid is None or not self._puppet.is_healthy():
            LOG.debug("Skipping coverage dump (target is not in a good state)")
            return
        # If at this point, the browser is in a good state, i.e. no crashes
        # or hangs, so signal the browser to dump coverage.
        try:
            for child in Process(pid).children(recursive=True):
                LOG.debug("Sending SIGUSR1 to %d (child)", child.pid)
                try:
                    kill(child.pid, SIGUSR1)
                except OSError:
                    LOG.warning("Failed to send SIGUSR1 to pid %d", child.pid)
        except (AccessDenied, NoSuchProcess):  # pragma: no cover
            pass
        LOG.debug("Sending SIGUSR1 to %d (parent)", pid)
        try:
            kill(pid, SIGUSR1)
        except OSError:
            LOG.warning("Failed to send SIGUSR1 to pid %d", pid)
        start_time = time()
        gcda_found = False
        delay = 0.1
        # wait for processes to write .gcno files
        # this should typically take less than 1 second
        while True:
            for proc in process_iter(attrs=["pid", "ppid", "open_files"]):
                # check if proc is the target or child process
                if pid in (proc.info["pid"], proc.info["ppid"]):
                    if proc.info["open_files"] is None:
                        continue
                    if any(x.path.endswith(".gcda") for x in proc.info["open_files"]):
                        gcda_found = True
                        # get the pid of the process that has the file open
                        gcda_open = proc.info["pid"]
                        break
            else:
                gcda_open = None
            elapsed = time() - start_time
            if gcda_found:
                if gcda_open is None:
                    # success
                    LOG.debug("gcda dump took %0.2fs", elapsed)
                    break
                if elapsed >= timeout:
                    # timeout failure
                    LOG.warning(
                        "gcda file open by pid %d after %0.2fs", gcda_open, elapsed
                    )
                    try:
                        kill(gcda_open, SIGABRT)
                    except OSError:
                        pass
                    sleep(1)
                    self.close()
                    break
                if delay < 1.0:
                    # increase delay to a maximum of 1 second
                    delay = min(1.0, delay + 0.1)
            elif elapsed >= 3:
                # assume we missed the process writing .gcno files
                LOG.warning("No gcda files seen after %0.2fs", elapsed)
                break
            if not self._puppet.is_healthy():
                LOG.warning("Browser failure during dump_coverage()")
                break
            sleep(delay)

    def launch(self, location, env_mod=None):
        # setup environment
        env_mod = dict(env_mod or [])
        # do not allow network connections to non local endpoints
        env_mod["MOZ_DISABLE_NONLOCAL_CONNECTIONS"] = "1"
        env_mod["MOZ_CRASHREPORTER_SHUTDOWN"] = "1"
        try:
            self._puppet.launch(
                self.binary,
                launch_timeout=self.launch_timeout,
                location=location,
                log_limit=self.log_limit,
                memory_limit=self.memory_limit,
                prefs_js=self.prefs,
                extension=self.extension,
                env_mod=env_mod,
            )
        except LaunchError as exc:
            LOG.error("FFPuppet LaunchError: %s", str(exc))
            self.close()
            if isinstance(exc, BrowserTimeoutError):
                raise TargetLaunchTimeout(str(exc)) from None
            raise TargetLaunchError(str(exc), self.create_report()) from None

    def log_size(self):
        return self._puppet.log_length("stderr") + self._puppet.log_length("stdout")

    @property
    def prefs(self):
        if self._prefs is None:
            # generate temporary prefs.js
            for prefs_template in PrefPicker.templates():
                if prefs_template.endswith("browser-fuzzing.yml"):
                    LOG.debug("using prefpicker template %r", prefs_template)
                    tmp_fd, self._prefs = mkstemp(
                        prefix="prefs_", suffix=".js", dir=grz_tmp()
                    )
                    close(tmp_fd)
                    PrefPicker.load_template(prefs_template).create_prefsjs(self._prefs)
                    LOG.debug("generated prefs.js %r", self._prefs)
                    self._remove_prefs = True
                    break
            else:  # pragma: no cover
                raise TargetError("Failed to generate prefs.js")
        return self._prefs

    @prefs.setter
    def prefs(self, prefs_file):
        if self._remove_prefs and self._prefs and isfile(self._prefs):
            unlink(self._prefs)
        if prefs_file is None:
            self._prefs = None
            self._remove_prefs = True
        elif isfile(prefs_file):
            self._prefs = abspath(prefs_file)
            self._remove_prefs = False
        else:
            raise TargetError("Missing prefs.js file %r" % (prefs_file,))

    def save_logs(self, *args, **kwargs):
        self._puppet.save_logs(*args, **kwargs)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Sapphire HTTP server job
"""
from collections import defaultdict, namedtuple
from logging import getLogger
from mimetypes import guess_type
from os import walk
from os.path import abspath, isdir, isfile
from os.path import join as pathjoin
from os.path import normpath, relpath, splitext
from queue import Queue
from threading import Event, Lock

from .server_map import Resource

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


# job status codes
SERVED_ALL = 0  # all expected requests for required files have been received
SERVED_NONE = 1  # no requests for required files have been received
SERVED_REQUEST = 2  # some requests for required files have been received
SERVED_TIMEOUT = 3  # timeout occurred


Tracker = namedtuple("Tracker", "files lock")


class Job:
    # MIME_MAP is used to support new or uncommon mime types.
    # Definitions in here take priority over mimetypes.guess_type().
    MIME_MAP = {
        ".avif": "image/avif",
        ".bmp": "image/bmp",
        ".ico": "image/x-icon",
        ".wave": "audio/x-wav",
        ".webp": "image/webp",
    }

    __slots__ = (
        "_complete",
        "_pending",
        "_served",
        "auto_close",
        "accepting",
        "base_path",
        "exceptions",
        "forever",
        "initial_queue_size",
        "server_map",
        "worker_complete",
    )

    def __init__(
        self,
        base_path,
        auto_close=-1,
        forever=False,
        optional_files=None,
        server_map=None,
    ):
        self._complete = Event()
        self._pending = Tracker(files=set(), lock=Lock())
        self._served = Tracker(files=defaultdict(int), lock=Lock())
        self.accepting = Event()
        self.accepting.set()
        self.auto_close = auto_close
        self.base_path = abspath(base_path)  # wwwroot
        self.exceptions = Queue()
        self.forever = forever
        self.initial_queue_size = 0
        self.server_map = server_map
        self.worker_complete = Event()
        self._build_queue(optional_files)

    def _build_queue(self, optional_files):
        # build file list to track files that must be served
        # this is intended to only be called once by __init__()
        for d_name, _, filenames in walk(self.base_path, followlinks=False):
            for f_name in filenames:
                file_path = pathjoin(d_name, f_name)
                location = relpath(file_path, start=self.base_path).replace("\\", "/")
                # do not add optional files to queue of required files
                if optional_files and location in optional_files:
                    LOG.debug("optional: %r", location)
                    continue
                file_path = abspath(file_path)
                if "?" in file_path:
                    LOG.warning(
                        "Cannot add files with '?' in path. Skipping %r", file_path
                    )
                    continue
                self._pending.files.add(file_path)
                LOG.debug("required: %r", location)
        # if nothing was found check if the path exists
        if not self._pending.files and not isdir(self.base_path):
            raise OSError("%r does not exist" % (self.base_path,))
        if self.server_map:
            for redirect, resource in self.server_map.redirect.items():
                if resource.required:
                    self._pending.files.add(redirect)
                LOG.debug(
                    "%s: %r -> %r",
                    "required" if resource.required else "optional",
                    redirect,
                    resource.target,
                )
            for dyn_resp, resource in self.server_map.dynamic.items():
                if resource.required:
                    self._pending.files.add(dyn_resp)
                    LOG.debug("%s: %r -> %r", "required", dyn_resp, resource.target)
        self.initial_queue_size = len(self._pending.files)
        LOG.debug("%d files required to serve", self.initial_queue_size)

    @classmethod
    def lookup_mime(cls, url):
        mime = cls.MIME_MAP.get(splitext(url)[-1].lower())
        if mime is None:
            # default to "application/octet-stream"
            mime = guess_type(url)[0] or "application/octet-stream"
        return mime

    def check_request(self, request):
        if "?" in request:
            request = request.split("?", 1)[0]
        to_serve = normpath(pathjoin(self.base_path, request))
        if "\x00" not in to_serve and isfile(to_serve):
            res = Resource(Resource.URL_FILE, to_serve, mime=self.lookup_mime(to_serve))
            with self._pending.lock:
                res.required = to_serve in self._pending.files
            return res
        if self.server_map is not None:
            if request in self.server_map.redirect:
                return self.server_map.redirect[request]
            if request in self.server_map.dynamic:
                return self.server_map.dynamic[request]
            # collect possible include matches
            includes = tuple(
                x for x in self.server_map.include if request.startswith(x)
            )
            if includes:
                LOG.debug("potential include matches %r", includes)
                # attempt to find match
                url = request
                while True:
                    if url in includes:
                        LOG.debug("found include match %r", url)
                        location = (
                            request.split(url, 1)[-1].lstrip("/") if url else request
                        )
                        # check location points to something
                        if location:
                            target = pathjoin(
                                self.server_map.include[url].target, location
                            )
                            # if the mapping url is empty check the file exists
                            if url or isfile(target):
                                mime = self.server_map.include[url].mime
                                if mime is None:
                                    mime = self.lookup_mime(to_serve)
                                return Resource(
                                    Resource.URL_INCLUDE,
                                    normpath(target),
                                    mime=mime,
                                    required=self.server_map.include[url].required,
                                )
                    if "/" in url:
                        url = url.rsplit("/", 1)[0]
                    elif url:
                        # try empty mount point
                        url = ""
                    else:
                        # include does not exist
                        break
        return None

    def finish(self):
        self._complete.set()

    def increment_served(self, target):
        # update list of served files
        with self._served.lock:
            self._served.files[target] += 1

    def is_complete(self, wait=None):
        if wait is not None:
            return self._complete.wait(wait)
        return self._complete.is_set()

    def is_forbidden(self, target_file):
        target_file = abspath(target_file)
        # check if target_file lives somewhere in wwwroot
        if not target_file.startswith(self.base_path):
            if self.server_map:
                for resources in self.server_map.include.values():
                    if target_file.startswith(resources.target):
                        return False  # this is a valid include path
            return True  # this is NOT a valid include path
        return False  # this is a valid path

    @property
    def pending(self):
        # number of pending files
        with self._pending.lock:
            return len(self._pending.files)

    def remove_pending(self, file_name):
        # return True when all file have been removed
        with self._pending.lock:
            if self._pending.files:
                self._pending.files.discard(file_name)
            return not self._pending.files

    @property
    def served(self):
        # served files
        # files served from www root will have a path relative to www root
        # include files will have an absolute path
        with self._served.lock:
            # make a copy of what is available (maybe a copy not necessary?)
            served = tuple(self._served.files.keys())
        for fname in served:
            if fname.startswith(self.base_path):
                # file is in www root
                yield relpath(fname, self.base_path)
            else:
                # include file
                yield fname

    @property
    def status(self):
        with self._pending.lock:
            queue_size = len(self._pending.files)
        if queue_size == 0:
            return SERVED_ALL
        if queue_size < self.initial_queue_size:
            return SERVED_REQUEST
        return SERVED_NONE



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger

from ..common.fuzzmanager import Bucket
from ..main import configure_logging
from ..session import Session
from .args import ReduceFuzzManagerIDQualityArgs
from .crash import main as crash_main

LOG = getLogger(__name__)


def main(args):
    """CLI for `grizzly.reduce.bucket`.

    Arguments:
        args (argparse.Namespace): Result from `ReduceArgs.parse_args`.

    Returns:
        int: 0 for success. non-0 indicates a problem.
    """
    configure_logging(args.log_level)
    LOG.info("Trying all crashes in bucket %d until one reduces", args.input)

    # ensure --tool is reset for each call to grizzly.reduce.crash.main()
    orig_tool = args.tool

    # if no crashes in bucket, return success
    result = Session.EXIT_SUCCESS

    bucket = Bucket(args.input)
    try:
        if args.sig is None:
            args.sig = str(bucket.signature_path())

        for crash in bucket.iter_crashes(args.quality):
            args.input = crash.crash_id
            args.tool = orig_tool

            # call grizzly.reduce.crash
            result = crash_main(args)
            if result in {Session.EXIT_SUCCESS, Session.EXIT_ABORT}:
                break

    finally:
        bucket.cleanup()
    return result


if __name__ == "__main__":
    raise SystemExit(main(ReduceFuzzManagerIDQualityArgs().parse_args()))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABCMeta, abstractmethod, abstractproperty
from logging import getLogger
from os.path import isfile
from threading import Lock

__all__ = ("Target", "TargetError", "TargetLaunchError")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]

LOG = getLogger(__name__)


class TargetError(Exception):
    """Raised by Target"""


class TargetLaunchError(TargetError):
    """Raised if a failure during launch occurs"""

    def __init__(self, message, report):
        super().__init__(message)
        self.report = report


class TargetLaunchTimeout(TargetError):
    """Raised if the target does not launch within the defined amount of time"""


class Target(metaclass=ABCMeta):
    RESULT_NONE = 0
    RESULT_FAILURE = 1
    RESULT_IGNORED = 2

    __slots__ = (
        "_lock",
        "_monitor",
        "_prefs",
        "binary",
        "extension",
        "launch_timeout",
        "log_limit",
        "memory_limit",
    )

    def __init__(self, binary, extension, launch_timeout, log_limit, memory_limit):
        assert log_limit >= 0
        assert memory_limit >= 0
        assert binary is not None and isfile(binary)
        self._lock = Lock()
        self._monitor = None
        self._prefs = None
        self.binary = binary
        self.extension = extension
        self.launch_timeout = max(launch_timeout, 300)
        self.log_limit = log_limit
        self.memory_limit = memory_limit

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.cleanup()

    def add_abort_token(self, _token):  # pylint: disable=no-self-use
        LOG.warning("add_abort_token() not implemented!")

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractproperty
    def closed(self):
        pass

    @abstractmethod
    def create_report(self, is_hang=False):
        pass

    @abstractmethod
    def detect_failure(self, ignored):
        pass

    def dump_coverage(self, _timeout=0):  # pylint: disable=no-self-use
        LOG.warning("dump_coverage() is not supported!")

    @abstractmethod
    def handle_hang(self, ignore_idle=True):
        pass

    # TODO: move to monitor?
    def is_idle(self, _threshold):  # pylint: disable=no-self-use
        LOG.debug("Target.is_idle() not implemented! returning False")
        return False

    @abstractmethod
    def launch(self, _location, _env_mod=None):
        pass

    def log_size(self):  # pylint: disable=no-self-use
        LOG.debug("log_size() not implemented! returning 0")
        return 0

    @abstractproperty
    def monitor(self):
        pass

    # TODO: better meta file handling
    @abstractproperty
    def prefs(self):
        pass

    def reverse(self, remote, local):
        # remote->device, local->desktop
        pass

    @abstractmethod
    def save_logs(self, *args, **kwargs):
        pass



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import DEBUG, basicConfig
from os import getenv, makedirs
from os.path import join as pathjoin
from re import split as resplit
from tempfile import gettempdir

__all__ = (
    "ConfigError",
    "configure_logging",
    "grz_tmp",
    "sanitizer_opts",
    "TIMEOUT_DELAY",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


# TIMEOUT_DELAY is added to the test time limit to create the default timeout
TIMEOUT_DELAY = 15


class ConfigError(Exception):
    """Raised to indicate invalid configuration a state"""

    def __init__(self, message, exit_code):
        super().__init__(message)
        self.exit_code = exit_code


def configure_logging(log_level):
    """Configure log output level and formatting.

    Args:
        log_level (int): Set log level.

    Returns:
        None
    """
    # allow force enabling log_level via environment
    if getenv("DEBUG", "0").lower() in ("1", "true"):
        log_level = DEBUG
    if log_level == DEBUG:
        date_fmt = None
        log_fmt = "%(asctime)s %(levelname).1s %(name)s | %(message)s"
    else:
        date_fmt = "%Y-%m-%d %H:%M:%S"
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt=date_fmt, level=log_level)


def grz_tmp(*subdir):
    path = pathjoin(gettempdir(), "grizzly", *subdir)
    makedirs(path, exist_ok=True)
    return path


def sanitizer_opts(env_data):
    """Parse the values defined in given *SAN_OPTIONS environment variable.
    For example "ASAN_OPTIONS=debug=false:log_path='/test/file.log'"
    would return {"debug": "false", "log_path": "'/test/file.log'"}

    Args:
        env_var (str): *SAN_OPTIONS environment variable to parse.

    Returns:
        dict: Sanitized values from environment.
    """
    opts = dict()
    for opt in resplit(r":(?![\\|/])", env_data):
        if not opt:
            continue
        key, val = opt.split("=")
        opts[key] = val
    return opts



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from shutil import rmtree
from tempfile import mkdtemp
from time import sleep, time

from sapphire import SERVED_TIMEOUT

from ..target import TargetLaunchError, TargetLaunchTimeout
from .utils import grz_tmp

__all__ = ("Runner", "RunResult")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class _IdleChecker:
    """_IdleChecker is used to help determine if the target is hung (actively using CPU)
    or if it has not made expected the HTTP requests for other reasons (idle).
    This will allow the framework to move on without interrupting execution of long
    running test cases.
    This is not perfect! It is to be used AFTER the test case timeout (initial_delay)
    has elapsed.
    """

    __slots__ = ("_check_cb", "_init_delay", "_poll_delay", "_threshold", "_next_poll")

    def __init__(self, check_cb, threshold, initial_delay, poll_delay=1):
        assert callable(check_cb)
        assert initial_delay >= 0
        assert poll_delay >= 0
        assert 100 > threshold >= 0
        self._check_cb = check_cb  # callback used to check if target is idle
        self._init_delay = initial_delay  # time to wait before the initial idle poll
        self._poll_delay = poll_delay  # time to wait between subsequent polls
        self._threshold = threshold  # CPU usage threshold
        self._next_poll = None

    def is_idle(self):
        """Check the target idle callback. This is throttled by '_next_poll'
        which specifies the time at which the next callback call is allowed.

        Args:
            None

        Returns:
            bool: True if the target idle callback returned True otherwise False
        """
        assert self._next_poll is not None, "schedule_poll() must be called first"
        now = time()
        if now >= self._next_poll:
            if self._check_cb(self._threshold):
                return True
            self.schedule_poll(now=now)
        return False

    def schedule_poll(self, initial=False, now=None):
        """Update `_next_poll`.

        Args:
            initial (int): Use `_init_delay` to schedule next poll
            now (int): time in seconds

        Returns:
            None
        """
        if now is None:
            now = time()
        if initial:
            self._next_poll = now + self._init_delay
        else:
            self._next_poll = now + self._poll_delay


class Runner:
    __slots__ = (
        "_close_delay",
        "_idle",
        "_relaunch",
        "_server",
        "_target",
        "_tests_run",
    )

    def __init__(
        self, server, target, close_delay=30, idle_threshold=0, idle_delay=0, relaunch=1
    ):
        self._close_delay = close_delay
        if idle_threshold > 0:
            assert idle_delay > 0
            LOG.debug("using idle check, th %d, delay %ds", idle_threshold, idle_delay)
            self._idle = _IdleChecker(target.is_idle, idle_threshold, idle_delay)
        else:
            self._idle = None
        assert close_delay > 0
        assert relaunch > 0
        self._relaunch = relaunch  # tests to run before relaunching target
        self._server = server  # a sapphire instance to serve the test case
        self._target = target  # target to run test case
        self._tests_run = 0  # number of tests run since target (re)launched

    def launch(self, location, env_mod=None, max_retries=3, retry_delay=0):
        """Launch a target and open `location`.

        Args:
            location (str): URL to open via Target.
            env_mod (dict): Environment modifications.
            max_retries (int): Number of retries to perform before re-raising
                               TargetLaunchTimeout.
            retry_delay (int): Time in seconds to wait between retries.

        Returns:
            None
        """
        assert self._server is not None
        assert self._target is not None
        assert max_retries >= 0
        assert retry_delay >= 0
        self._server.clear_backlog()
        LOG.debug("launching target (timeout %ds)", self._target.launch_timeout)
        for retries in reversed(range(max_retries)):
            try:
                self._target.launch(location, env_mod=env_mod)
            except TargetLaunchError as exc:
                # This is likely due to a bad build or environment configuration.
                if retries:
                    LOG.warning("Failure detected during launch (retries %d)", retries)
                    exc.report.cleanup()
                    sleep(retry_delay)
                    continue
                raise
            except TargetLaunchTimeout:
                # A TargetLaunchTimeout likely has nothing to do with Grizzly but is
                # seen frequently on machines under a high load. After multiple
                # consecutive timeouts something is likely wrong so raise.
                if retries:
                    LOG.warning("Timeout detected during launch (retries %d)", retries)
                    sleep(retry_delay)
                    continue
                raise
            break
        self._tests_run = 0

    @staticmethod
    def location(srv_path, srv_port, close_after=None, time_limit=None):
        """Build a valid URL to pass to a browser.

        Args:
            srv_path (str): Path segment of the URL
            srv_port (int): Server listening port
            close_after (int): Harness argument.
            time_limit (int): Harness argument.

        Returns:
            str: A valid URL.
        """
        location = "http://127.0.0.1:%d/%s" % (srv_port, srv_path.lstrip("/"))
        # set harness related arguments
        args = []
        if close_after is not None:
            assert close_after >= 0
            args.append("close_after=%d" % (close_after,))
        if time_limit is not None:
            assert time_limit >= 0
            args.append("time_limit=%d" % (time_limit * 1000,))
        if args:
            return "?".join([location, "&".join(args)])
        return location

    def run(
        self,
        ignore,
        server_map,
        testcase,
        test_path=None,
        coverage=False,
        wait_for_callback=False,
    ):
        """Serve a testcase and monitor the target for results.

        Args:
            ignore (list): List of failure types to ignore.
            server_map (sapphire.ServerMap): A ServerMap.
            testcase (grizzly.TestCase): The test case that will be served.
            test_path (str): Location of test case data on the filesystem.
            coverage (bool): Trigger coverage dump.
            wait_for_callback: (bool): Use `_keep_waiting()` to indicate when
                                       framework should move on.

        Returns:
            RunResult: Files served, status and timeout flag from the run.
        """
        if self._idle is not None:
            self._idle.schedule_poll(initial=True)
        if self._tests_run == self._relaunch - 1:
            # overwrite instead of replace 'grz_next_test' for consistency
            server_map.set_redirect("grz_next_test", "grz_empty", required=True)
            server_map.set_dynamic_response("grz_empty", lambda: b"", required=True)
        try:
            # unpack test case
            if test_path is None:
                wwwdir = mkdtemp(prefix="test_", dir=grz_tmp("serve"))
                testcase.dump(wwwdir)
            else:
                wwwdir = test_path
            # serve the test case
            serve_start = time()
            server_status, served = self._server.serve_path(
                wwwdir,
                continue_cb=self._keep_waiting,
                forever=wait_for_callback,
                optional_files=tuple(testcase.optional),
                server_map=server_map,
            )
            duration = time() - serve_start
        finally:
            # remove temporary files
            if test_path is None:
                rmtree(wwwdir)
        result = RunResult(served, duration, timeout=server_status == SERVED_TIMEOUT)
        result.attempted = testcase.landing_page in result.served
        result.initial = self._tests_run == 0
        # TODO: fix calling TestCase.add_batch() for multi-test replay
        # add all include files that were served
        for url, resource in server_map.include.items():
            testcase.add_batch(resource.target, result.served, prefix=url)
        if result.timeout:
            LOG.debug("timeout detected")
            if self._target.handle_hang(ignore_idle=True) or "timeout" in ignore:
                result.status = RunResult.IGNORED
            server_map.dynamic.pop("grz_empty", None)
        if result.attempted:
            self._tests_run += 1
            if coverage and not result.timeout:
                # dump_coverage() should be called before detect_failure()
                # to help catch any coverage related issues.
                self._target.dump_coverage()
            # relaunch check
            if self._tests_run >= self._relaunch and not result.timeout:
                assert self._tests_run == self._relaunch
                server_map.dynamic.pop("grz_empty", None)
                LOG.debug("relaunch/shutdown limit hit")
                # ideally all browser tabs should be closed at this point
                # and the browser should exit on its own
                # NOTE: this will take the full duration if target.is_idle()
                # is not implemented
                for close_delay in range(max(int(self._close_delay / 0.5), 1)):
                    if not self._target.monitor.is_healthy():
                        break
                    # wait 3 seconds (6 passes) before attempting idle exit
                    if close_delay > 5 and self._target.is_idle(10):
                        # NOTE: this will always trigger on systems where the
                        # browser does not exit when the last window is closed
                        LOG.debug("target idle")
                        break
                    # delay to help catch shutdown related crashes, LSan, etc.
                    # debugger and different builds can slow shutdown
                    sleep(0.5)
                else:
                    LOG.debug("target.close() required")
                self._target.close()
        else:
            # something is wrong so close the target
            # previous iteration put target in a bad state?
            LOG.debug("landing page %r not served!", testcase.landing_page)
            self._target.close()
        # detect failure
        if result.status is None:
            failure_detected = self._target.detect_failure(ignore)
            if failure_detected == self._target.RESULT_FAILURE:
                result.status = RunResult.FAILED
            elif failure_detected == self._target.RESULT_IGNORED:
                result.status = RunResult.IGNORED
        return result

    def _keep_waiting(self):
        """Callback used by the server to determine if it should continue to
        wait for the requests from the target.

        Args:
            None

        Returns:
            bool: Continue to serve the test case.
        """
        if self._idle is not None and self._idle.is_idle():
            LOG.debug("idle target detected")
            return False
        return self._target.monitor.is_healthy()


class RunResult:
    """A RunResult holds result details from a call to Runner.run().

    Attributes:
        attempted (bool): Test landing page (entry point) was requested.
        duration (float): Time spent waiting for test contents to be served.
        initial (bool): Target was (re)launched prior to run attempt.
        served (tuple(str)): Files that were served.
        status (int): Result status of test.
        timeout (bool): A timeout occurred waiting for test to complete.
    """

    FAILED = 1
    IGNORED = 2

    __slots__ = ("attempted", "duration", "initial", "served", "status", "timeout")

    def __init__(self, served, duration, status=None, timeout=False):
        self.attempted = False
        self.duration = duration
        self.initial = False
        self.served = served
        self.status = status
        self.timeout = timeout



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer testcase list strategy definition."""

from logging import getLogger

from ...common.storage import TestCase
from . import Strategy

LOG = getLogger(__name__)


class MinimizeTestcaseList(Strategy):
    """Try removing testcases from a list of sequential testcases (eg. Grizzly result
    cache). The strategy favours testcases at the tail of the list, so for a list of
    five testcases:
                            testcases
                    0       1 2 3 4 5
        iteration   1         2 3 4 5
                    2       1   3 4 5
                    3       1 2   4 5
                    4       1 2 3   5
                    5       1 2 3 4
    """

    name = "list"

    def __init__(self, testcases):
        """Initialize strategy instance.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase)):
                List of testcases to reduce. The object does not take ownership of the
                testcases.
        """
        super().__init__(testcases)
        self._current_feedback = None
        self._current_served = None

    def update(self, success, served=None):
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success (bool): Whether or not the last reduction was acceptable.
            served (list(list(str))): The list of served files for each testcase in the
                                      last reduction.

        Returns:
            None
        """
        assert self._current_feedback is None
        assert self._current_served is None
        self._current_feedback = success
        self._current_served = served

    def __iter__(self):
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each testcase set yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            list(grizzly.common.storage.TestCase): list of testcases with reduction
                                                   applied
        """
        assert self._current_feedback is None
        idx = 0
        testcases = None
        try:
            testcases = TestCase.load(str(self._testcase_root), True)
            n_testcases = len(testcases)
            # indicates that self._testcase_root contains changes that haven't been
            # yielded (if iteration ends, changes would be lost)
            testcase_root_dirty = False
            while True:
                if n_testcases <= 1:
                    LOG.info(
                        "Testcase list has length %d, not enough to reduce!",
                        n_testcases,
                    )
                    break
                if idx >= n_testcases:
                    LOG.info("Attempted to remove every single testcase")
                    break
                # try removing the testcase at idx
                if testcases is None:
                    testcases = TestCase.load(str(self._testcase_root), True)
                    assert n_testcases == len(testcases)
                testcases.pop(idx).cleanup()
                yield testcases
                testcases = None  # caller owns testcases now
                assert self._current_feedback is not None, "no feedback received!"

                if self._current_feedback:
                    testcase_root_dirty = False
                    LOG.info(
                        "Removing testcase %d/%d was successful!", idx + 1, n_testcases
                    )
                    testcases = TestCase.load(str(self._testcase_root), True)
                    try:
                        # remove the actual testcase we were reducing
                        testcases.pop(idx).cleanup()
                        if testcases and self._current_served is not None:
                            testcase_root_dirty = self.purge_unserved(
                                testcases, self._current_served
                            )
                        else:
                            self.dump_testcases(testcases, recreate_tcroot=True)
                    finally:
                        for testcase in testcases:
                            testcase.cleanup()
                    testcases = TestCase.load(str(self._testcase_root), True)
                    n_testcases = len(testcases)
                else:
                    LOG.info("No result without testcase %d/%d", idx + 1, n_testcases)
                    idx += 1
                # reset
                self._current_feedback = None
                self._current_served = None
            if testcase_root_dirty:
                # purging unserved files enabled us to exit early from the loop.
                # need to yield once more to set this trimmed version to the current
                # best in ReduceManager
                testcases = TestCase.load(str(self._testcase_root), True)
                LOG.info("[%s] final iteration triggered by purge_optional", self.name)
                yield testcases
                testcases = None  # caller owns testcases now
                assert (
                    self._current_feedback
                ), "Purging unserved files broke the testcase."
        finally:
            if testcases is not None:
                for testcase in testcases:
                    testcase.cleanup()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.replay.main
"""
from shutil import rmtree

from pytest import mark

from sapphire import SERVED_ALL

from ..common import Report, TestCase, TestCaseLoadFailure
from ..session import Session
from ..target import Target, TargetLaunchError, TargetLaunchTimeout
from .replay import ReplayManager
from .test_replay import _fake_save_logs


def test_main_01(mocker, tmp_path):
    """test ReplayManager.main()"""
    # This is a typical scenario - a test that reproduces results ~50% of the time.
    # Of the four attempts only the first and third will 'reproduce' the result
    # and the forth attempt should be skipped.
    # mock Sapphire.serve_path only
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path", autospec=True
    )
    serve_path.return_value = (
        SERVED_ALL,
        ["test.html"],
    )  # passed to mocked Target.detect_failure
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin")
    target = mocker.Mock(spec=Target, binary="bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.side_effect = (
        Target.RESULT_FAILURE,
        Target.RESULT_NONE,
        Target.RESULT_FAILURE,
    )
    target.save_logs = _fake_save_logs
    load_target.return_value.return_value = target
    # setup args
    log_path = tmp_path / "logs"
    (tmp_path / "test.html").touch()
    (tmp_path / "prefs.js").touch()
    (tmp_path / "sig.json").write_bytes(
        b'{"symptoms": [{"type": "crashAddress", "address": "0"}]}'
    )
    args = mocker.Mock(
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=str(tmp_path / "test.html"),
        logs=str(log_path),
        min_crashes=2,
        no_harness=True,
        prefs=str(tmp_path / "prefs.js"),
        relaunch=1,
        repeat=4,
        rr=False,
        sig=str(tmp_path / "sig.json"),
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Session.EXIT_SUCCESS
    assert target.reverse.call_count == 1
    assert target.launch.call_count == 3
    assert target.detect_failure.call_count == 3
    assert serve_path.call_count == 3
    assert load_target.call_count == 1
    assert target.close.call_count == 4
    assert target.cleanup.call_count == 1
    assert log_path.is_dir()
    assert any(log_path.glob("reports/*/log_asan_blah.txt"))
    assert any(log_path.glob("reports/*/log_stderr.txt"))
    assert any(log_path.glob("reports/*/log_stdout.txt"))


def test_main_02(mocker, tmp_path):
    """test ReplayManager.main() - no repro"""
    # mock Sapphire.serve_path only
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path", autospec=True
    )
    serve_path.return_value = (
        SERVED_ALL,
        ["test.html"],
    )  # passed to mocked Target.detect_failure
    # setup Target
    load_target = mocker.patch("grizzly.replay.replay.load_plugin")
    target = mocker.Mock(spec=Target, binary="bin", launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    load_target.return_value.return_value = target
    # setup args
    (tmp_path / "test.html").touch()
    (tmp_path / "prefs.js").touch()
    args = mocker.Mock(
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["fake", "timeout"],
        input=str(tmp_path / "test.html"),
        min_crashes=2,
        no_harness=True,
        prefs=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
        valgrind=False,
    )
    assert ReplayManager.main(args) == Session.EXIT_FAILURE
    assert target.detect_failure.call_count == 1
    assert target.close.call_count == 2
    assert target.cleanup.call_count == 1


def test_main_03(mocker):
    """test ReplayManager.main() error cases"""
    fake_sig = mocker.patch("grizzly.replay.replay.CrashSignature", autospec=True)
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    fake_load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    fake_tc = mocker.patch("grizzly.replay.replay.TestCase", autospec=True)
    # setup args
    args = mocker.Mock(
        ignore=list(),
        input="test",
        min_crashes=1,
        no_harenss=True,
        prefs=None,
        relaunch=1,
        repeat=1,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
    )
    # user abort
    fake_load_target.side_effect = KeyboardInterrupt
    # coverage
    args.rr = True
    args.valgrind = False
    assert ReplayManager.main(args) == Session.EXIT_ABORT
    fake_load_target.reset_mock()
    # invalid test case
    fake_tc.load.side_effect = TestCaseLoadFailure
    # coverage
    args.rr = False
    args.valgrind = True
    assert ReplayManager.main(args) == Session.EXIT_ERROR
    assert fake_load_target.call_count == 0
    # no test cases
    args.valgrind = False
    fake_tc.load.side_effect = None
    fake_tc.load.return_value = list()
    assert ReplayManager.main(args) == Session.EXIT_ERROR
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # multiple test cases with --no-harness
    fake_tc.load.return_value = [mocker.Mock(hang=False), mocker.Mock(hang=False)]
    assert ReplayManager.main(args) == Session.EXIT_ARGS
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # signature required replaying hang
    fake_tc.load.return_value = [mocker.Mock(hang=True)]
    assert ReplayManager.main(args) == Session.EXIT_ERROR
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()
    # can't ignore timeout replaying hang
    args.ignore = ["timeout"]
    args.sig = "sig"
    fake_tc.load.return_value = [mocker.Mock(hang=True)]
    assert ReplayManager.main(args) == Session.EXIT_ERROR
    assert fake_sig.fromFile.call_count == 1
    assert fake_load_target.call_count == 0
    fake_load_target.reset_mock()


def test_main_04(mocker, tmp_path):
    """test ReplayManager.main() target exceptions"""
    mocker.patch("grizzly.replay.replay.FuzzManagerReporter", autospec=True)
    mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    mocker.patch("grizzly.replay.replay.TestCase", autospec=True)
    target = mocker.Mock(spec=Target, launch_timeout=30)
    load_target = mocker.patch("grizzly.replay.replay.load_plugin", autospec=True)
    load_target.return_value.return_value = target
    fake_tmp = tmp_path / "grz_tmp"
    fake_tmp.mkdir()
    mocker.patch(
        "grizzly.replay.replay.grz_tmp", autospec=True, return_value=str(fake_tmp)
    )
    # setup args
    args = mocker.Mock(
        ignore=list(),
        input="test",
        min_crashes=1,
        no_harenss=True,
        prefs=None,
        relaunch=1,
        repeat=1,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=None,
    )
    # target launch error
    fake_logs = tmp_path / "fake_report"
    fake_logs.mkdir()
    report = mocker.Mock(spec=Report, prefix="fake_report", path=str(fake_logs))
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run",
        side_effect=TargetLaunchError("", report),
    )
    assert ReplayManager.main(args) == Session.EXIT_LAUNCH_FAILURE
    assert not fake_logs.is_dir()
    assert "fake_report_logs" in (x.name for x in fake_tmp.iterdir())
    # target launch timeout
    mocker.patch(
        "grizzly.replay.replay.ReplayManager.run", side_effect=TargetLaunchTimeout
    )
    assert ReplayManager.main(args) == Session.EXIT_LAUNCH_FAILURE


def test_main_05(mocker, tmp_path):
    """test ReplayManager.main() loading/generating prefs.js"""
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path", autospec=True
    )
    serve_path.return_value = (
        None,
        ["test.html"],
    )  # passed to mocked Target.detect_failure
    # setup Target
    target = mocker.Mock(spec=Target, binary="bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    load_target = mocker.patch("grizzly.replay.replay.load_plugin")
    load_target.return_value.return_value = target
    # setup args
    args = mocker.Mock(
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=list(),
        min_crashes=1,
        no_harness=True,
        relaunch=1,
        repeat=1,
        sig=None,
        test_index=None,
        time_limit=1,
        timeout=None,
    )
    log_path = tmp_path / "logs"
    args.logs = str(log_path)
    input_path = tmp_path / "input"
    input_path.mkdir()
    # build a test case
    entry_point = input_path / "test.html"
    entry_point.touch()
    with TestCase("test.html", None, "test-adapter") as src:
        src.add_from_file(str(entry_point))
        src.dump(str(input_path), include_details=True)
    args.input = str(input_path)

    # test no specified prefs.js
    args.prefs = None
    assert ReplayManager.main(args) == Session.EXIT_SUCCESS
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_path.call_count == 1
    assert log_path.is_dir()
    assert not any(log_path.glob("**/prefs.js"))

    target.reset_mock()
    serve_path.reset_mock()
    rmtree(str(log_path), ignore_errors=True)

    # test included prefs.js
    (input_path / "prefs.js").write_bytes(b"included")
    assert ReplayManager.main(args) == Session.EXIT_SUCCESS
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_path.call_count == 1
    assert log_path.is_dir()
    prefs = next(log_path.glob("**/prefs.js"))
    assert prefs.read_bytes() == b"included"

    target.reset_mock()
    serve_path.reset_mock()
    rmtree(str(log_path), ignore_errors=True)

    # test specified prefs.js
    (tmp_path / "prefs.js").write_bytes(b"specified")
    args.prefs = str(tmp_path / "prefs.js")
    assert ReplayManager.main(args) == Session.EXIT_SUCCESS
    assert target.launch.call_count == 1
    assert target.detect_failure.call_count == 1
    assert serve_path.call_count == 1
    assert log_path.is_dir()
    prefs = next(log_path.glob("**/prefs.js"))
    assert prefs.read_bytes() == b"specified"


@mark.parametrize(
    "arg_timelimit, arg_timeout, test_timelimit, result",
    [
        # use default test time limit and timeout values (test missing time limit)
        (None, None, None, Session.EXIT_FAILURE),
        # use min test time limit and default timeout values
        (None, None, 1, Session.EXIT_FAILURE),
        # set test time limit
        (10, None, None, Session.EXIT_FAILURE),
        # set both test time limit and timeout to the same value
        (10, 10, None, Session.EXIT_FAILURE),
        # set timeout greater than test time limit
        (10, 11, None, Session.EXIT_FAILURE),
        # set test time limit greater than timeout
        (11, 10, None, Session.EXIT_ARGS),
    ],
)
def test_main_06(mocker, tmp_path, arg_timelimit, arg_timeout, test_timelimit, result):
    """test ReplayManager.main() - test time limit and timeout"""
    # mock Sapphire.serve_path only
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    serve_path = mocker.patch(
        "grizzly.replay.replay.Sapphire.serve_path", autospec=True
    )
    serve_path.return_value = (
        SERVED_ALL,
        ["test.html"],
    )  # passed to mocked Target.detect_failure
    # setup Target
    target = mocker.Mock(spec=Target, binary="bin", launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    load_target = mocker.patch("grizzly.replay.replay.load_plugin")
    load_target.return_value.return_value = target
    # create test to load
    test = TestCase("test.html", None, None)
    test_file = tmp_path / "test.html"
    test_file.write_text("test")
    test.add_from_file(str(test_file))
    replay_path = tmp_path / "test"
    replay_path.mkdir()
    test.time_limit = test_timelimit
    test.dump(str(replay_path), include_details=True)
    # setup args
    (tmp_path / "prefs.js").touch()
    args = mocker.Mock(
        fuzzmanager=False,
        idle_delay=0,
        idle_threshold=0,
        ignore=["timeout"],
        input=str(replay_path),
        min_crashes=2,
        no_harness=True,
        prefs=None,
        relaunch=1,
        repeat=1,
        rr=False,
        sig=None,
        test_index=None,
        time_limit=arg_timelimit,
        timeout=arg_timeout,
        valgrind=False,
    )
    assert ReplayManager.main(args) == result



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import DEBUG, INFO

from pytest import mark

from .utils import configure_logging, grz_tmp, sanitizer_opts


def test_grz_tmp_01(mocker, tmp_path):
    """test grz_tmp()"""
    mocker.patch(
        "grizzly.common.utils.gettempdir", autospec=True, return_value=str(tmp_path)
    )
    # create temp path
    path = grz_tmp()
    assert path == str(tmp_path / "grizzly")
    assert (tmp_path / "grizzly").is_dir()
    # create temp path (exists)
    path = grz_tmp()
    assert path == str(tmp_path / "grizzly")
    # create temp path with sub directory
    path = grz_tmp("test1", "test2")
    assert path == str(tmp_path / "grizzly" / "test1" / "test2")
    assert (tmp_path / "grizzly" / "test1" / "test2").is_dir()


@mark.parametrize(
    "env, log_level",
    [
        # default log level
        ("0", INFO),
        # debug log level
        ("0", DEBUG),
        # enable debug log level via env
        ("1", INFO),
        # enable debug log level via env
        ("TRUE", INFO),
    ],
)
def test_configure_logging_01(mocker, env, log_level):
    """test configure_logging()"""
    config = mocker.patch("grizzly.common.utils.basicConfig", autospec=True)
    mocker.patch("grizzly.common.utils.getenv", autospec=True, return_value=env)
    configure_logging(log_level)
    assert config.call_count == 1
    if env != "0":
        assert config.call_args[-1]["level"] == DEBUG
    else:
        assert config.call_args[-1]["level"] == log_level


@mark.parametrize(
    "to_parse, expected",
    [
        # test empty string
        ("", {}),
        # test single value
        ("test_value=true", {"test_value": "true"}),
        # test multiple values
        ("a=1:b=-2:C=3", {"a": "1", "b": "-2", "C": "3"}),
        # path parsing
        (
            "p1='z:/a':p2='x:\\a.1':p3='/test/path/':p4='':p5=\"x:/a.a\"",
            {
                "p1": "'z:/a'",
                "p2": "'x:\\a.1'",
                "p3": "'/test/path/'",
                "p4": "''",
                "p5": '"x:/a.a"',
            },
        ),
        # test platform specific parsing
        (
            "bar=1:file='%s':foo=2" % (__file__,),
            {"bar": "1", "file": "'%s'" % (__file__,), "foo": "2"},
        ),
    ],
)
def test_sanitizer_opts_01(to_parse, expected):
    """test sanitizer_opts()"""
    parsed = sanitizer_opts(to_parse)
    assert len(parsed) == len(expected)
    for key in expected:
        assert expected[key] == parsed[key]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.replay
"""
from itertools import cycle
from os.path import join as pathjoin

from pytest import mark, raises

from sapphire import SERVED_ALL, SERVED_NONE, SERVED_REQUEST, SERVED_TIMEOUT, Sapphire

from ..common import Report, Status, TestCase, TestCaseLoadFailure
from ..target import Target
from .replay import ReplayManager, ReplayResult


def _fake_save_logs(result_logs, _meta=False):
    """write fake log data to disk"""
    with open(pathjoin(result_logs, "log_stderr.txt"), "w") as log_fp:
        log_fp.write("STDERR log\n")
    with open(pathjoin(result_logs, "log_stdout.txt"), "w") as log_fp:
        log_fp.write("STDOUT log\n")
    with open(pathjoin(result_logs, "log_asan_blah.txt"), "w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: ")
        log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19\n")


def test_replay_01(mocker):
    """test ReplayManager.cleanup()"""
    replay = ReplayManager(
        [], mocker.Mock(spec=Sapphire), mocker.Mock(spec=Target), [mocker.Mock()]
    )
    replay.status = mocker.Mock(spec=Status)
    replay.cleanup()
    assert replay.status.cleanup.call_count == 1


def test_replay_02(mocker, tmp_path):
    """test ReplayManager.run() - no repro"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=True, launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    target.monitor.is_healthy.return_value = False
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        with ReplayManager([], server, target, use_harness=True, relaunch=1) as replay:
            assert not replay.run([testcase], 10)
            assert replay.signature is None
            assert replay.status.ignored == 0
            assert replay.status.iteration == 1
            assert replay.status.results == 0
            assert target.monitor.is_healthy.call_count == 1
            assert target.close.call_count == 2
        assert not any(tmp_path.iterdir())


def test_replay_03(mocker):
    """test ReplayManager.run() - no repro - with repeats"""
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=False, launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    target.monitor.is_healthy.return_value = False
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        with ReplayManager(
            [], server, target, use_harness=True, relaunch=100
        ) as replay:
            assert not replay.run([testcase], 10, repeat=10, min_results=1)
            assert replay.signature is None
            assert replay.status.ignored == 0
            assert replay.status.iteration == 10
            assert replay.status.results == 0
            assert target.handle_hang.call_count == 0
            assert target.monitor.is_healthy.call_count == 1
            assert target.close.call_count == 2


def test_replay_04(mocker):
    """test ReplayManager.run() - exit - skip shutdown in runner"""
    # this will make runner appear to have just relaunched the target
    # and skip the expected shutdown
    mocker.patch(
        "grizzly.common.runner.Runner._tests_run",
        new_callable=mocker.PropertyMock,
        return_value=0,
    )
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=False, launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        with ReplayManager(
            [], server, target, use_harness=True, relaunch=100
        ) as replay:
            assert not replay.run([testcase], 10, repeat=10, min_results=1)
            assert replay.status.ignored == 0
            assert replay.status.iteration == 10
            assert replay.status.results == 0
            assert target.handle_hang.call_count == 0
            assert target.monitor.is_healthy.call_count == 0
            assert target.close.call_count == 1


def test_replay_05(mocker, tmp_path):
    """test ReplayManager.run() - successful repro"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    served = ["index.html"]
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, served)
    target = mocker.Mock(spec=Target, binary="C:\\fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        with ReplayManager([], server, target, relaunch=10) as replay:
            assert replay.signature is None
            results = replay.run([testcase], 10)
            assert replay.signature is not None
            assert replay.status.ignored == 0
            assert replay.status.iteration == 1
            assert replay.status.results == 1
            assert target.handle_hang.call_count == 0
            assert target.monitor.is_healthy.call_count == 1
            assert target.close.call_count == 2
        assert len(results) == 1
        assert results[0].count == 1
        assert results[0].expected
        assert results[0].report
        assert len(results[0].served) == 1
        assert results[0].served[0] == served
        assert len(results[0].durations) == 1
        results[0].report.cleanup()
        assert not any(tmp_path.iterdir())


def test_replay_06(mocker, tmp_path):
    """test ReplayManager.run() - error - landing page not requested"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, binary="bin", closed=True, launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    # test target unresponsive
    target.detect_failure.return_value = Target.RESULT_NONE
    server.serve_path.return_value = (SERVED_NONE, [])
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(testcases, 10, repeat=1)
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 0
        # target.close() called once in runner and once by ReplayManager.run()
        assert target.close.call_count == 2
    target.reset_mock()
    # test target crashed
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.save_logs = _fake_save_logs
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(testcases, 10, repeat=1)
        assert replay.status.ignored == 1
        assert replay.status.iteration == 1
        assert replay.status.results == 0
        assert replay._signature is None
        # target.close() called once in runner and once by ReplayManager.run()
        assert target.close.call_count == 2
    assert len(results) == 1
    assert results[0].count == 1
    assert not results[0].expected


def test_replay_07(mocker, tmp_path):
    """test ReplayManager.run()
    delayed failure - following test landing page not requested"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    target = mocker.Mock(spec=Target, binary="bin", launch_timeout=30)
    type(target).closed = mocker.PropertyMock(side_effect=(True, False, True))
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.side_effect = (Target.RESULT_NONE, Target.RESULT_FAILURE)
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    server.serve_path.side_effect = (
        (SERVED_ALL, ["index.html"]),
        (SERVED_REQUEST, ["x"]),
    )
    with ReplayManager([], server, target, use_harness=True, relaunch=10) as replay:
        assert replay.run(testcases, 10, repeat=2)
        assert replay.status.ignored == 0
        assert replay.status.iteration == 2
        assert replay.status.results == 1
        # target.close() called once in runner and once by ReplayManager.run()
        assert target.close.call_count == 2


def test_replay_08(mocker, tmp_path):
    """test ReplayManager.run() - ignored"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=True, launch_timeout=30)
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.detect_failure.return_value = Target.RESULT_IGNORED
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(testcases, 10)
        assert target.monitor.is_healthy.call_count == 1
        assert target.close.call_count == 2
        assert replay.status.ignored == 1
        assert replay.status.iteration == 1
        assert replay.status.results == 0
    assert not any(tmp_path.iterdir())


def test_replay_09(mocker, tmp_path):
    """test ReplayManager.run() - early exit"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, binary="path/fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_IGNORED = Target.RESULT_IGNORED
    target.RESULT_NONE = Target.RESULT_NONE
    target.save_logs = _fake_save_logs
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    # early failure
    target.detect_failure.side_effect = (
        Target.RESULT_FAILURE,
        Target.RESULT_IGNORED,
        Target.RESULT_NONE,
    )
    target.monitor.is_healthy.side_effect = (False, False, True, False)
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(testcases, 10, repeat=4, min_results=3)
        assert target.close.call_count == 4
        assert replay.status.iteration == 3
        assert replay.status.results == 1
        assert replay.status.ignored == 1
    # early success
    target.reset_mock()
    target.detect_failure.side_effect = (
        Target.RESULT_FAILURE,
        Target.RESULT_IGNORED,
        Target.RESULT_FAILURE,
    )
    target.monitor.is_healthy.side_effect = (False, False, False)
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(testcases, 10, repeat=4, min_results=2)
        assert target.close.call_count == 4
        assert replay.status.iteration == 3
        assert replay.status.results == 2
        assert replay.status.ignored == 1
    assert len(results) == 1
    assert sum(x.count for x in results) == 2
    target.reset_mock()
    # ignore early failure (perform all repeats)
    target.detect_failure.return_value = Target.RESULT_NONE
    target.detect_failure.side_effect = None
    target.monitor.is_healthy.side_effect = None
    target.monitor.is_healthy.return_value = True
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(testcases, 10, repeat=4, min_results=4, exit_early=False)
        assert target.close.call_count == 5
        assert replay.status.iteration == 4
        assert replay.status.results == 0
        assert replay.status.ignored == 0
    target.reset_mock()
    # ignore early success (perform all repeats)
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(testcases, 10, repeat=4, min_results=1, exit_early=False)
        assert target.close.call_count == 5
        assert replay.status.iteration == 4
        assert replay.status.results == 4
        assert replay.status.ignored == 0
    assert len(results) == 1
    assert sum(x.count for x in results) == 4


def test_replay_10(mocker, tmp_path):
    """test ReplayManager.run() - test signatures - fail to meet minimum"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_0 = mocker.Mock(spec=Report)
    report_0.crash_info.createShortSignature.return_value = "No crash detected"
    report_1 = mocker.Mock(
        spec=Report, crash_hash="hash1", major="0123abcd", minor="01239999"
    )
    report_1.crash_info.createShortSignature.return_value = "[@ test1]"
    report_2 = mocker.Mock(
        spec=Report, crash_hash="hash2", major="0123abcd", minor="abcd9876"
    )
    report_2.crash_info.createShortSignature.return_value = "[@ test2]"
    report_3 = mocker.Mock(
        spec=Report, crash_hash="hash2", major="0123abcd", minor="abcd9876"
    )
    report_3.crash_info.createShortSignature.return_value = "[@ test2]"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0, report_1, report_2, report_3)
    fake_report.calc_hash.return_value = "bucketHASH"
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    signature = mocker.Mock()
    signature.matches.side_effect = (True, False, False)
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager(
        [], server, target, signature=signature, use_harness=False
    ) as replay:
        results = replay.run(testcases, 10, repeat=4, min_results=2)
        assert target.close.call_count == 5
        assert replay.signature == signature
        assert replay.status.iteration == 4
        assert replay.status.results == 1
        assert replay.status.ignored == 2
    assert fake_report.call_count == 4
    assert len(results) == 1
    assert not results[0].expected
    assert results[0].count == 2
    assert report_0.cleanup.call_count == 1
    assert report_1.cleanup.call_count == 1
    assert report_2.cleanup.call_count == 0
    assert report_3.cleanup.call_count == 1
    assert signature.matches.call_count == 3


def test_replay_11(mocker, tmp_path):
    """test ReplayManager.run() - test signatures - multiple matches"""
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_0 = mocker.Mock(
        spec=Report, crash_hash="hash1", major="0123abcd", minor="01239999"
    )
    report_0.crash_info.createShortSignature.return_value = "[@ test1]"
    report_1 = mocker.Mock(
        spec=Report, crash_hash="hash2", major="0123abcd", minor="abcd9876"
    )
    report_1.crash_info.createShortSignature.return_value = "[@ test2]"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0, report_1)
    fake_report.calc_hash.return_value = "bucketHASH"
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["a.html"])
    signature = mocker.Mock()
    signature.matches.side_effect = (True, True)
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="a.html", optional=[])
    ]
    with ReplayManager(
        [], server, target, signature=signature, use_harness=False
    ) as replay:
        results = replay.run(testcases, 10, repeat=2, min_results=2)
        assert target.close.call_count == 3
        assert replay.signature == signature
        assert replay.status.iteration == 2
        assert replay.status.results == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 2
    assert len(results) == 1
    assert results[0].expected
    assert results[0].count == 2
    assert report_0.cleanup.call_count == 0
    assert report_1.cleanup.call_count == 1
    assert signature.matches.call_count == 2


def test_replay_12(mocker, tmp_path):
    """test ReplayManager.run() - any crash - success"""
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_0 = mocker.Mock(spec=Report)
    report_0.crash_info.createShortSignature.return_value = "No crash detected"
    report_1 = mocker.Mock(
        spec=Report, crash_hash="hash1", major="0123abcd", minor="01239999"
    )
    report_1.crash_info.createShortSignature.return_value = "[@ test1]"
    report_2 = mocker.Mock(
        spec=Report, crash_hash="hash2", major="0123abcd", minor="abcd9876"
    )
    report_2.crash_info.createShortSignature.return_value = "[@ test2]"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0, report_1, report_2)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager([], server, target, any_crash=True, use_harness=False) as replay:
        results = replay.run(testcases, 10, repeat=3, min_results=2)
        assert target.close.call_count == 4
        assert replay.signature is None
        assert replay.status.iteration == 3
        assert replay.status.results == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 3
    assert len(results) == 2
    assert all(x.expected for x in results)
    assert sum(x.count for x in results if x.expected) == 2
    assert report_0.cleanup.call_count == 1
    assert report_1.cleanup.call_count == 0
    assert report_2.cleanup.call_count == 0


def test_replay_13(mocker, tmp_path):
    """test ReplayManager.run() - any crash - fail to meet minimum"""
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_0 = mocker.Mock(spec=Report)
    report_0.crash_info.createShortSignature.return_value = "No crash detected"
    report_1 = mocker.Mock(
        spec=Report, crash_hash="hash1", major="0123abcd", minor="01239999"
    )
    report_1.crash_info.createShortSignature.return_value = "[@ test1]"
    report_2 = mocker.Mock(
        spec=Report, crash_hash="hash2", major="0123abcd", minor="abcd9876"
    )
    report_2.crash_info.createShortSignature.return_value = "[@ test2]"
    report_3 = mocker.Mock(spec=Report)
    report_3.crash_info.createShortSignature.return_value = "No crash detected"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0, report_1, report_2, report_3)
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager([], server, target, any_crash=True) as replay:
        assert not replay.run(testcases, 10, repeat=4, min_results=3)
        assert target.close.call_count == 5
        assert replay.signature is None
        assert replay.status.iteration == 4
        assert replay.status.results == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 4
    assert report_0.cleanup.call_count == 1
    assert report_1.cleanup.call_count == 1
    assert report_2.cleanup.call_count == 1
    assert report_3.cleanup.call_count == 1


def test_replay_14(mocker, tmp_path):
    """test ReplayManager.run() - no signature - use first crash"""
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_1 = mocker.Mock(spec=Report, crash_hash="hash1", major="0123", minor="9999")
    report_1.crash_info.createShortSignature.return_value = "[@ test1]"
    auto_sig = mocker.Mock()
    auto_sig.matches.side_effect = (True, False, True)
    report_1.crash_signature = auto_sig
    report_2 = mocker.Mock(spec=Report, crash_hash="hash2", major="abcd", minor="9876")
    report_2.crash_info.createShortSignature.return_value = "[@ test2]"
    report_3 = mocker.Mock(spec=Report, crash_hash="hash1", major="0123", minor="9999")
    report_3.crash_info.createShortSignature.return_value = "[@ test1]"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_1, report_2, report_3)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(testcases, 10, repeat=3, min_results=2)
        assert target.close.call_count == 4
        assert replay.signature == auto_sig
        assert replay.status.iteration == 3
        assert replay.status.results == 2
        assert replay.status.ignored == 1
    assert fake_report.call_count == 3
    assert len(results) == 2
    assert sum(x.expected for x in results) == 1
    assert sum(x.count for x in results if x.expected) == 2
    assert report_1.cleanup.call_count == 0
    assert report_2.cleanup.call_count == 0
    assert report_3.cleanup.call_count == 1


def test_replay_15(mocker, tmp_path):
    """test ReplayManager.run() - unexpected exception"""
    mocker.patch(
        "grizzly.replay.replay.mkdtemp", autospec=True, return_value=str(tmp_path)
    )
    report_0 = mocker.Mock(
        spec=Report, crash_hash="hash1", major="0123abcd", minor="01239999"
    )
    report_0.crash_info.createShortSignature.return_value = "[@ test1]"
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0,)
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.side_effect = ((SERVED_ALL, ["index.html"]), KeyboardInterrupt)
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[])
    ]
    with ReplayManager(
        [], server, target, any_crash=True, use_harness=True, relaunch=2
    ) as replay:
        with raises(KeyboardInterrupt):
            replay.run(testcases, 10, repeat=3, min_results=2)
        assert replay.signature is None
        assert replay.status.iteration == 2
        assert replay.status.results == 1
        assert replay.status.ignored == 0
    assert target.close.call_count == 1
    assert target.monitor.is_healthy.call_count == 0
    assert fake_report.call_count == 1
    assert report_0.cleanup.call_count == 1


def test_replay_16(mocker):
    """test ReplayManager.run() - multiple TestCases - no repro"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=True, launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
    ]
    with ReplayManager([], server, target, use_harness=True) as replay:
        assert not replay.run(testcases, 10)
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 0
    assert target.close.call_count == 2
    assert all(x.dump.call_count == 1 for x in testcases)


def test_replay_17(mocker):
    """test ReplayManager.run() - multiple TestCases - no repro - with repeats"""
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, launch_timeout=30)
    # test relaunch < repeat
    type(target).closed = mocker.PropertyMock(side_effect=cycle([True, False]))
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    target.monitor.is_healthy.return_value = False
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="index.html", optional=[]),
    ]
    with ReplayManager([], server, target, use_harness=True, relaunch=2) as replay:
        assert not replay.run(testcases, 10, repeat=10)
        assert server.serve_path.call_count == 30
        assert target.close.call_count == 6
        assert target.launch.call_count == 5
        assert replay.status.ignored == 0
        assert replay.status.iteration == 10
        assert replay.status.results == 0
    assert target.monitor.is_healthy.call_count == 5
    assert all(x.dump.call_count == 1 for x in testcases)


def test_replay_18(mocker, tmp_path):
    """test ReplayManager.run() - multiple TestCases - successful repro"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.side_effect = (
        (SERVED_ALL, ["a.html"]),
        (SERVED_ALL, ["b.html"]),
        (SERVED_ALL, ["c.html"]),
    )
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.side_effect = (
        Target.RESULT_NONE,
        Target.RESULT_NONE,
        Target.RESULT_FAILURE,
    )
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    testcases = [
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="a.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="b.html", optional=[]),
        mocker.Mock(spec=TestCase, env_vars=[], landing_page="c.html", optional=[]),
    ]
    with ReplayManager([], server, target, use_harness=True) as replay:
        results = replay.run(testcases, 30)
        assert target.close.call_count == 2
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results == 1
    assert len(results) == 1
    assert len(results[0].served) == len(testcases)
    assert results[0].served[0][0] == "a.html"
    assert results[0].served[1][0] == "b.html"
    assert results[0].served[2][0] == "c.html"
    assert len(results[0].durations) == len(testcases)
    assert all(x.dump.call_count == 1 for x in testcases)


def test_replay_19(mocker, tmp_path):
    """test ReplayManager.run() - multiple calls"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_ALL, ["index.html"])
    target = mocker.Mock(spec=Target, closed=True, launch_timeout=30)
    target.RESULT_NONE = Target.RESULT_NONE
    target.detect_failure.return_value = Target.RESULT_NONE
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        with ReplayManager([], server, target, use_harness=True) as replay:
            assert not replay.run([testcase], 30)
            assert replay.status.iteration == 1
            assert not replay.run([testcase], 30)
            assert replay.status.iteration == 1
            assert not replay.run([testcase], 30)
            assert replay.status.iteration == 1
    assert server.serve_path.call_count == 3


def test_replay_20(mocker, tmp_path):
    """test ReplayManager.report_to_filesystem()"""
    # no reports
    ReplayManager.report_to_filesystem(str(tmp_path), [])
    assert not any(tmp_path.iterdir())
    # with reports and tests
    (tmp_path / "report_expected").mkdir()
    result0 = mocker.Mock(
        ReplayResult, count=1, durations=[1], expected=True, served=[]
    )
    result0.report = mocker.Mock(
        spec=Report, path=str(tmp_path / "report_expected"), prefix="expected"
    )
    (tmp_path / "report_other1").mkdir()
    result1 = mocker.Mock(
        ReplayResult, count=1, durations=[1], expected=False, served=None
    )
    result1.report = mocker.Mock(
        spec=Report, path=str(tmp_path / "report_other1"), prefix="other1"
    )
    (tmp_path / "report_other2").mkdir()
    result2 = mocker.Mock(
        ReplayResult, count=1, durations=[1], expected=False, served=None
    )
    result2.report = mocker.Mock(
        spec=Report, path=str(tmp_path / "report_other2"), prefix="other2"
    )
    test = mocker.Mock(spec=TestCase)
    path = tmp_path / "dest"
    ReplayManager.report_to_filesystem(
        str(path), [result0, result1, result2], tests=[test]
    )
    assert test.dump.call_count == 3  # called once per report
    assert not (tmp_path / "report_expected").is_dir()
    assert not (tmp_path / "report_other1").is_dir()
    assert not (tmp_path / "report_other2").is_dir()
    assert path.is_dir()
    assert (path / "reports").is_dir()
    assert (path / "reports" / "expected_logs").is_dir()
    assert (path / "other_reports").is_dir()
    assert (path / "other_reports" / "other1_logs").is_dir()
    assert (path / "other_reports" / "other2_logs").is_dir()
    # with reports and no tests
    (tmp_path / "report_expected").mkdir()
    result0.reset_mock()
    path = tmp_path / "dest2"
    ReplayManager.report_to_filesystem(str(path), [result0])
    assert not (tmp_path / "report_expected").is_dir()
    assert path.is_dir()
    assert (path / "reports" / "expected_logs").is_dir()


def test_replay_21(mocker, tmp_path):
    """test ReplayManager.load_testcases()"""
    fake_load = mocker.patch("grizzly.replay.replay.TestCase.load")
    test0 = mocker.Mock(spec=TestCase)
    test1 = mocker.Mock(spec=TestCase, landing_page="x.html")
    test2 = mocker.Mock(spec=TestCase)
    # failure
    fake_load.return_value = ()
    with raises(TestCaseLoadFailure, match="Failed to load TestCases"):
        ReplayManager.load_testcases(str(tmp_path), False)
    # success
    fake_load.return_value = [test0, test1]
    tests = ReplayManager.load_testcases(str(tmp_path), False)
    assert len(tests) == 2
    assert tests[0].cleanup.call_count == 0
    assert tests[1].cleanup.call_count == 0
    # success select
    fake_load.return_value = [test0, test1, test2, mocker.Mock(spec=TestCase)]
    tests = ReplayManager.load_testcases(str(tmp_path), False, subset=[1, 3])
    assert len(tests) == 2
    assert tests[0].landing_page == "x.html"
    assert test0.cleanup.call_count == 1
    assert test1.cleanup.call_count == 0
    assert test2.cleanup.call_count == 1
    test0.reset_mock()
    test2.reset_mock()
    # select (first and last) with invalid input
    fake_load.return_value = [test0, test1, test2]
    tests = ReplayManager.load_testcases(str(tmp_path), False, subset=[0, 10, -10, -1])
    assert len(tests) == 2
    assert test0.cleanup.call_count == 0
    assert test1.cleanup.call_count == 1
    assert test2.cleanup.call_count == 0


@mark.parametrize(
    "is_hang, use_sig, match_sig, ignored, results",
    [
        # reproduce expected hang
        (True, True, True, 0, 1),
        # expected hang (signature, no match)
        (True, True, False, 1, 0),
        # unexpected hang (signature, match)
        (False, True, True, 0, 1),
        # unexpected hang (signature, no match)
        (False, True, False, 1, 0),
        # unexpected hang (no signature)
        (False, False, False, 1, 0),
    ],
)
def test_replay_22(mocker, tmp_path, is_hang, use_sig, match_sig, ignored, results):
    """test ReplayManager.run() - detect hangs"""
    mocker.patch("grizzly.replay.replay.grz_tmp", return_value=str(tmp_path))
    served = ["index.html"]
    server = mocker.Mock(spec=Sapphire, port=0x1337, timeout=10)
    server.serve_path.return_value = (SERVED_TIMEOUT, served)
    if use_sig:
        signature = mocker.Mock()
        signature.matches.return_value = match_sig
        signature.rawSignature = "fakesig"
    else:
        signature = None
    target = mocker.Mock(spec=Target, binary="fake_bin", launch_timeout=30)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.detect_failure.return_value = Target.RESULT_FAILURE
    target.handle_hang.return_value = False
    target.save_logs = _fake_save_logs
    with TestCase("index.html", "redirect.html", "test-adapter") as testcase:
        testcase.hang = is_hang
        with ReplayManager(
            [], server, target, signature=signature, relaunch=10
        ) as replay:
            found = replay.run([testcase], 10, expect_hang=is_hang)
            assert replay.status.iteration == 1
            assert replay.status.ignored == ignored
            assert replay.status.results == results
            assert target.handle_hang.call_count == 1
            assert target.monitor.is_healthy.call_count == 0
            assert target.close.call_count == 1
        assert len(found) == 1
        assert found[0].count == 1
        assert found[0].expected == results
        assert found[0].report
        assert found[0].report.is_hang
        assert len(found[0].served) == 1
        assert found[0].served[0] == served
        assert len(found[0].durations) == 1
        assert testcase.hang == is_hang
        found[0].report.cleanup()
        assert not any(tmp_path.iterdir())



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""Common unit test fixtures for `grizzly.reduce`.
"""

import pytest


@pytest.fixture
def tmp_path_fm_config(tmp_path, mocker):
    """Ensure fm config is always read from tmp_path so ~/.fuzzmanagerconf
    can't be used by accident.
    """
    mocker.patch(
        "grizzly.reduce.core.FuzzManagerReporter.FM_CONFIG",
        new=str(tmp_path / ".fuzzmanagerconf"),
    )
    (tmp_path / ".fuzzmanagerconf").touch()


@pytest.fixture
def reporter_sequential_strftime(mocker):
    """Make `strftime` in `FilesystemReporter` return sequential values.
    This ensures ever report gets a unique folder and won't overwrite another.
    """
    prefix = mocker.patch("grizzly.common.reporter.strftime")

    def report_prefix(_):
        return "%04d" % (prefix.call_count,)

    prefix.side_effect = report_prefix
